from ase import units
import numpy as np
import re
import warnings

__all__ = ["MDUnits"]


class MDUnits:
    """
    Basic conversion factors to atomic units used internally:
        fs2atu (time): femtoseconds to atomic time units
        eV2Ha (energy): electron Volt to Hartree
        d2amu (mass): Dalton to atomic mass units
        angs2bohr (length): Angstrom to Bohr
        auforces2aseforces (forces): Hartee per Bohr to electron Volt per Angstrom

    Definitions for constants:
        kB: Boltzmann constant in units of Hartree per Kelvin.
        hbar: Reduced Planck constant in atomic units.
    """

    # Internal units (MD internal -> ASE internal)
    energy_unit = units.kJ / units.mol
    length_unit = units.nm
    mass_unit = 1.0  # 1 Dalton in ASE reference frame

    # Derived units (MD internal -> ASE internal)
    time_unit = length_unit * np.sqrt(mass_unit / energy_unit)

    # General utility units for conversion
    fs2internal = units.fs / time_unit
    da2internal = 1.0 / mass_unit
    angs2internal = units.Angstrom / length_unit
    bar2internal = 1e5 * units.Pascal / (energy_unit / length_unit ** 3)

    # Constants (internal frame)
    kB = units.kB / energy_unit  # Always uses Kelvin
    hbar = (
        units._hbar * (units.J * units.s) / (energy_unit * time_unit)
    )  # hbar is given in J/s by ASE

    # For spectra
    h_bar2icm = hbar * 100 * units._c * units._aut

    # Conversion units use when reading in MD inputs
    # These always convert to ASE frame of reference
    conversions = {
        "mol": units.mol,
        "kcal": units.kcal / energy_unit,
        "kj": units.kJ / energy_unit,
        "ev": units.eV / energy_unit,
        "hartree": units.Ha / energy_unit,
        "ha": units.Ha / energy_unit,
        "bohr": units.Bohr / length_unit,
        "angstrom": units.Angstrom / length_unit,
        "angs": units.Angstrom / length_unit,
        "a": units.Angstrom / length_unit,
        "nm": units.nm / length_unit,
        "fs": units.fs / time_unit,
        "s": units.s / time_unit,
        "aut": units._aut * units.s / time_unit,  # _aut is given in s
    }

    @staticmethod
    def parse_mdunit(unit):
        """
        Auxiliary functions, used to determine the conversion factor of position and force units between MD propagation
        and the provided ML Calculator. Allowed units are:
            mol, kcal, eV, Bohr, Angstrom, Hartree and all combinations using '/' thereof (e.g. kcal/mol/Angstrom).

        Args:
            unit (str/float): Unit to be used to convert forces from Calculator units to atomic units used in the MD.
                              Can be a str, which is converted to the corresponding numerical value or a float, which
                              is returned.

        Returns:
            float: Factor used for conversion in the Calculator.

        """
        warnings.warn(
            "Routine is deprecated, please use unit2internal, internal2unit or unit2unit instead.",
            DeprecationWarning,
        )
        if type(unit) == str:
            parts = unit.lower().replace(" ", "").split("/")
            scaling = 1.0
            for part in parts:
                if part not in MDUnits.conversions:
                    raise KeyError("Unrecognized unit {:s}".format(part))
                scaling *= MDUnits.conversions[part]
            return scaling
        else:
            return unit

    @staticmethod
    def _parse_unit(unit):
        if type(unit) == str:
            # If a string is given, split into parts.
            parts = re.split("(\W)", unit.lower())

            conversion = 1.0
            divide = False
            for part in parts:
                if part == "/":
                    divide = True
                elif part == "" or part == " ":
                    pass
                else:
                    if divide:
                        conversion /= MDUnits.conversions[part]
                        divide = False
                    else:
                        conversion *= MDUnits.conversions[part]
            return conversion
        else:
            # If unit is given as number, return number
            return unit

    @staticmethod
    def unit2internal(unit):
        conversion = MDUnits._parse_unit(unit)
        return conversion

    @staticmethod
    def internal2unit(unit):
        conversion = MDUnits._parse_unit(unit)
        return 1.0 / conversion

    @staticmethod
    def unit2unit(unit1, unit2):
        conversion1 = MDUnits.unit2internal(unit1)
        conversion2 = MDUnits.internal2unit(unit2)
        return conversion1 * conversion2



import schnetpack
from schnetpack import Properties

from schnetpack.md.calculators import MDCalculator
from schnetpack.md.calculators.ensemble_calculator import EnsembleCalculator
from schnetpack.md.neighbor_lists import SimpleNeighborList
import logging

from schnetpack.md.utils import MDUnits, check_triples_required
from schnetpack.environment import collect_atom_triples_batch


class SchnetPackCalculator(MDCalculator):
    """
    MD calculator for schnetpack models.

    Args:
        model (schnetpack.atomistic.AtomisticModel): Loaded schnetpack model.
        required_properties (list): List of properties to be computed by the calculator
        force_handle (str): String indicating the entry corresponding to the molecular forces
        position_conversion (float): Unit conversion for the length used in the model computing all properties. E.g. if
                             the model needs Angstrom, one has to provide the conversion factor converting from the
                             atomic units used internally (Bohr) to Angstrom: 0.529177...
        force_conversion (float): Conversion factor converting the forces returned by the used model back to atomic
                                  units (Hartree/Bohr).
        property_conversion (dict(float)): Optional dictionary of conversion factors for other properties predicted by
                                           the model. Only changes the units used for logging the various outputs.
        detach (bool): Detach property computation graph after every calculator call. Enabled by default. Should only
                       be disabled if one wants to e.g. compute derivatives over short trajectory snippets.
        neighbor_list (schnetpack.md.neighbor_list.MDNeighborList): Neighbor list object for determining which
                                                                    interatomic distances should be computed.
        cutoff (float): Cutoff radius for computing the neighbor interactions. If this is set to a negative number,
                        the cutoff is determined automatically based on the model (default=-1.0). Units are the distance
                        units used in the model.
        cutoff_shell (float): Second shell around the cutoff region. The neighbor lists only are recomputed when atoms move
                              a distance further than this shell (default=1 Angstrom).
    """

    def __init__(
        self,
        model,
        required_properties,
        force_handle,
        position_conversion="Angstrom",
        force_conversion="eV / Angstrom",
        property_conversion={},
        stress_handle=None,
        stress_conversion="eV / Angstrom / Angstrom / Angstrom",
        detach=True,
        neighbor_list=SimpleNeighborList,
        cutoff=-1.0,
        cutoff_shell=1.0,
        cutoff_lr=None,
    ):
        super(SchnetPackCalculator, self).__init__(
            required_properties,
            force_handle,
            position_conversion=position_conversion,
            force_conversion=force_conversion,
            stress_handle=stress_handle,
            stress_conversion=stress_conversion,
            property_conversion=property_conversion,
            detach=detach,
        )
        self.model = model

        # Check if atom triples need to be computed (e.g. for Behler functions)
        self.triples_required = check_triples_required(model)

        # If stress is required, activate stress computation in model
        if stress_handle is not None:
            self._activate_stress(stress_handle)

        self.neighbor_list = self._init_neighbor_list(
            neighbor_list, cutoff, cutoff_shell, cutoff_lr
        )

    def _activate_stress(self, stress_handle):
        schnetpack.utils.activate_stress_computation(self.model, stress=stress_handle)

    def calculate(self, system):
        """
        Main routine, generates a properly formatted input for the schnetpack model from the system, performs the
        computation and uses the results to update the system state.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.
        """
        # set model to evaluation mode to disable graph creation
        self.model.eval()

        inputs = self._generate_input(system)
        self.results = self.model(inputs)
        self._update_system(system)

    def _init_neighbor_list(self, neighbor_list, cutoff, cutoff_shell, cutoff_lr):
        """
        Function for properly setting up the neighbor list used for the SchNetPack calculator.
        This automatically checks, whether a proper cutoff has been provided and moves neighbor lists which support
        CUDA to the appropriate device.

        Args:
            neighbor_list (schnetpack.md.neighbor_lists.MDNeighborList.__init__): Uninitialized neighbor list class.
            cutoff (float): Cutoff radius for computing the neighbor interactions. If this is set to a negative number,
                            the cutoff is determined automatically based on the model (default=-1.0). Units are the
                            distance units used in the model.
            cutoff_shell (float): Second shell around the cutoff region. The neighbor lists only are recomputed when
                                  atoms move a distance further than this shell (default=1 Angstrom).

        Returns:
            schnetpack.md.neighbor_lists.MDNeighborList: Initialized neighbor list.
        """
        # Determine the cutoff for the neighbor list
        if cutoff is not None:
            if cutoff < 0.0:
                # Get cutoff automatically if given one is negative
                cutoff = self._get_model_cutoff()
                # cutoff *= self.position_conversion
            else:
                # Check whether set cutoff is reasonable
                try:
                    model_cutoff = self._get_model_cutoff()
                    # cutoff *= self.position_conversion
                    if cutoff < model_cutoff:
                        logging.warning(
                            f"Specified cutoff for neighbor list {cutoff} smaller than cutoff in model {model_cutoff}."
                        )
                except:
                    print("No cutoff found in model, using {:f}".format(cutoff))

            # Convert from model units to internal units
            cutoff /= self.position_conversion
            cutoff_shell /= self.position_conversion

        if isinstance(neighbor_list, schnetpack.md.neighbor_lists.TorchNeighborList):
            device = next(self.model.parameters()).device
        else:
            device = None

        if cutoff_lr is not None:
            cutoff_lr /= self.position_conversion
            neighbor_list = schnetpack.md.neighbor_lists.DualNeighborList(
                cutoff, cutoff_lr, neighbor_list, shell=cutoff_shell, device=device
            )
        else:
            neighbor_list = neighbor_list(cutoff, shell=cutoff_shell, device=device)

        return neighbor_list

    def _generate_input(self, system):
        """
        Function to extracts neighbor lists, atom_types, positions e.t.c. from the system and generate a properly
        formatted input for the schnetpack model.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.

        Returns:
            dict(torch.Tensor): Schnetpack inputs in dictionary format.
        """
        positions, atom_types, atom_masks, cells, pbc = self._get_system_molecules(
            system
        )

        inputs = {
            Properties.R: positions,
            Properties.Z: atom_types,
            Properties.atom_mask: atom_masks,
            Properties.cell: cells,
            Properties.pbc: pbc,
        }

        neighbors = self._get_system_neighbors(system)
        inputs.update(neighbors)

        return inputs

    def _get_system_neighbors(self, system):
        """
        Auxiliary function, which extracts neighbor lists formatted for schnetpack models from the system class.
        This is done by collapsing the replica and molecule dimension into one batch dimension.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.

        Returns:
            torch.LongTensor: (n_replicas*n_molecules) x n_atoms x (n_atoms-1) tensor holding the indices of all
                              neighbor atoms.
            torch.LongTensor: (n_replicas*n_molecules) x n_atoms x (n_atoms-1) binary tensor indicating padded
                              dimensions.
            torch.LongTensor: (n_replicas*n_molecules) x n_atoms x (n_atoms-1) x 3 tensor of the cell offset vectors
                              for periodic boundary conditions or cell simulations.
        """
        if self.neighbor_list is None:
            raise ValueError("System does not have neighbor list.")

        neighbor_list, neighbor_mask, offsets = self.neighbor_list.get_neighbors(system)

        neighbor_list, neighbor_mask, offsets = self._format_neighbors(
            neighbor_list,
            neighbor_mask,
            offsets,
            system.max_n_atoms,
            self.neighbor_list.max_neighbors,
        )
        neighbors = {
            Properties.neighbors: neighbor_list,
            Properties.neighbor_mask: neighbor_mask,
            Properties.cell_offset: offsets,
        }

        # Check if triples are required
        if self.triples_required:
            (
                nbh_j,
                nbh_k,
                offset_idx_j,
                offset_idx_k,
                pair_mask,
            ) = collect_atom_triples_batch(neighbor_list, neighbor_mask)
            neighbors[Properties.neighbor_pairs_j] = nbh_j
            neighbors[Properties.neighbor_pairs_k] = nbh_k
            neighbors[Properties.neighbor_offsets_j] = offset_idx_j
            neighbors[Properties.neighbor_offsets_k] = offset_idx_k
            neighbors[Properties.neighbor_pairs_mask] = pair_mask

        # Check if two cutoffs are present
        if isinstance(
            self.neighbor_list, schnetpack.md.neighbor_lists.DualNeighborList
        ):
            (
                neighbor_list_lr,
                neighbor_mask_lr,
                offsets_lr,
            ) = self.neighbor_list.get_neighbors_lr(system)

            neighbor_list_lr, neighbor_mask_lr, offsets_lr = self._format_neighbors(
                neighbor_list_lr,
                neighbor_mask_lr,
                offsets_lr,
                system.max_n_atoms,
                self.neighbor_list.max_neighbors_lr,
            )
            neighbors.update(
                {
                    Properties.neighbors_lr: neighbor_list_lr,
                    Properties.neighbor_mask_lr: neighbor_mask_lr,
                    Properties.cell_offset_lr: offsets_lr,
                }
            )

        return neighbors

    @staticmethod
    def _format_neighbors(
        neighbor_list, neighbor_mask, offsets, max_n_atoms, max_neighbors
    ):
        neighbor_list = neighbor_list.view(-1, max_n_atoms, max_neighbors)
        neighbor_mask = neighbor_mask.view(-1, max_n_atoms, max_neighbors)

        # Offsets need not be transformed to units, only cell
        if offsets is not None:
            offsets = offsets.view(-1, max_n_atoms, max_neighbors, 3)
        return neighbor_list, neighbor_mask, offsets

    def _get_model_cutoff(self):
        """
        Function to check the model passed to the calculator for already set cutoffs. This depends on the representation
        used, currently SchNet and wACSF models are supported.

        Returns:
            float: Model cutoff in model position units.
        """

        # Get representation
        if hasattr(self.model, "module"):
            representation = self.model.module.representation
        else:
            representation = self.model.representation

        # Check for different models, cutoff is set in different functions
        if isinstance(representation, schnetpack.representation.SchNet):
            model_cutoff = representation.interactions[0].cutoff_network.cutoff
        elif isinstance(representation, schnetpack.representation.BehlerSFBlock):
            model_cutoff = representation.cutoff.cutoff
        elif isinstance(representation, schnetpack.representation.StandardizeSF):
            model_cutoff = representation.SFBlock.cutoff.cutoff
        else:
            raise ValueError(
                "Unrecognized model representation {:s }for cutoff detection.".format(
                    representation
                )
            )

        # Convert from torch tensor and print out the detected cutoff
        model_cutoff = float(model_cutoff[0].cpu().numpy())
        logging.info("Detected cutoff radius of {:5.3f}...".format(model_cutoff))

        return model_cutoff


class EnsembleSchnetPackCalculator(EnsembleCalculator, SchnetPackCalculator):
    def __init__(
        self,
        models,
        required_properties,
        force_handle,
        position_conversion="Angstrom",
        force_conversion="eV / Angstrom",
        property_conversion={},
        stress_handle=None,
        stress_conversion="eV / Angstrom / Angstrom / Angstrom",
        detach=True,
        neighbor_list=SimpleNeighborList,
        cutoff=-1.0,
        cutoff_shell=1.0,
        cutoff_lr=None,
    ):
        self.models = models
        required_properties = self._update_required_properties(required_properties)

        super(EnsembleSchnetPackCalculator, self).__init__(
            models[0],
            required_properties=required_properties,
            force_handle=force_handle,
            position_conversion=position_conversion,
            force_conversion=force_conversion,
            property_conversion=property_conversion,
            stress_handle=stress_handle,
            stress_conversion=stress_conversion,
            detach=detach,
            neighbor_list=neighbor_list,
            cutoff=cutoff,
            cutoff_shell=cutoff_shell,
            cutoff_lr=cutoff_lr,
        )



import os
from ase.db import connect
from .atoms import AtomsData, AtomsDataError


__all__ = ["merge_datasets", "save_dataset"]


def merge_datasets(merged_dbpath, dbpaths, **mergedb_kwargs):

    if type(dbpaths) is dict:
        names = dbpaths.keys()
        dbpaths = dbpaths.values()
    else:
        names = [dbp.split("/")[-1].split(".")[0] for dbp in dbpaths]

    partitions = {}
    offset = 0

    partition_meta = {}
    with connect(merged_dbpath, use_lock_file=False) as dst:
        for name, dbp in zip(names, dbpaths):
            start = offset

            if name in partitions.keys():
                count = 1
                while name + "_" + str(count) in partitions.keys():
                    count += 1
                name = name + "_" + str(count)

            with connect(dbp) as src:
                length = src.count()
                end = offset + length
                partition_meta[name] = src.metadata

                for row in src.select():
                    at = row.toatoms()
                    dst.write(at, key_value_pairs=row.key_value_pairs, data=row.data)
            partitions[name] = (start, end)
            offset += length

    metadata = {"partition_meta": partition_meta, "partitions": partitions}
    dst.metadata = metadata

    return AtomsData(merged_dbpath, **mergedb_kwargs)


def save_dataset(dbpath, dataset, overwrite=False):
    """
    Write dataset instance to ase-db file.

    Args:
        dbpath (str): path to the new database
        dataset (spk.data.ConcatAtomsData or spk.data.AtomsDataSubset): dataset
            instance to be stored
        overwrite (bool): overwrite existing database

    """
    # check if path exists
    if os.path.exists(dbpath):
        if overwrite:
            os.remove(dbpath)
        raise AtomsDataError(
            "The selected dbpath does already exist. Set overwrite=True or change "
            "dbpath."
        )

    # build metadata
    metadata = dict()
    metadata["atomref"] = dataset.atomref
    metadata["available_properties"] = dataset.available_properties

    # write dataset
    with connect(dbpath) as conn:
        # write metadata
        conn.metadata = metadata
        # write datapoints
        for idx in range(len(dataset)):
            atms, data = dataset.get_properties(idx)
            # filter available properties
            data_clean = dict()
            for pname, prop in data.items():
                if pname.startswith("_"):
                    continue
                if pname in dataset.available_properties:
                    data_clean[pname] = prop

            conn.write(atms, data=data_clean)



"""
This module provides a ASE calculator class [#ase1]_ for SchNetPack models, as
well as a general Interface to all ASE calculation methods, such as geometry
optimisation, normal mode computation and molecular dynamics simulations.

References
----------
.. [#ase1] Larsen, Mortensen, Blomqvist, Castelli, Christensen, Dułak, Friis,
    Groves, Hammer, Hargus: The atomic simulation environment -- a Python
    library for working with atoms.
    Journal of Physics: Condensed Matter, 9, 27. 2017.
"""

import os

from schnetpack.data.atoms import AtomsConverter
from schnetpack.utils.spk_utils import DeprecationHelper
from schnetpack import Properties

from ase import units
from ase.calculators.calculator import Calculator, all_changes
from ase.io import read, write
from ase.io.trajectory import Trajectory
from ase.io.xyz import read_xyz, write_xyz
from ase.md import VelocityVerlet, Langevin, MDLogger
from ase.md.velocitydistribution import (
    MaxwellBoltzmannDistribution,
    Stationary,
    ZeroRotation,
)
from ase.optimize import QuasiNewton
from ase.vibrations import Vibrations

from schnetpack.md.utils import MDUnits

from schnetpack.environment import SimpleEnvironmentProvider


class SpkCalculatorError(Exception):
    pass


class SpkCalculator(Calculator):
    """
    ASE calculator for schnetpack machine learning models.

    Args:
        ml_model (schnetpack.AtomisticModel): Trained model for
            calculations
        device (str): select to run calculations on 'cuda' or 'cpu'
        collect_triples (bool): Set to True if angular features are needed,
            for example, while using 'wascf' models
        environment_provider (callable): Provides neighbor lists
        pair_provider (callable): Provides list of neighbor pairs. Only
            required if angular descriptors are used. Default is none.
        **kwargs: Additional arguments for basic ase calculator class
    """

    energy = Properties.energy
    forces = Properties.forces
    stress = Properties.stress
    implemented_properties = [energy, forces, stress]

    def __init__(
        self,
        model,
        device="cpu",
        collect_triples=False,
        environment_provider=SimpleEnvironmentProvider(),
        energy=None,
        forces=None,
        stress=None,
        energy_units="eV",
        forces_units="eV/Angstrom",
        stress_units="eV/Angstrom/Angstrom/Angstrom",
        **kwargs
    ):
        Calculator.__init__(self, **kwargs)

        self.model = model
        self.model.to(device)

        self.atoms_converter = AtomsConverter(
            environment_provider=environment_provider,
            collect_triples=collect_triples,
            device=device,
        )

        self.model_energy = energy
        self.model_forces = forces
        self.model_stress = stress

        # Convert to ASE internal units (energy=eV, length=A)
        self.energy_units = MDUnits.unit2unit(energy_units, "eV")
        self.forces_units = MDUnits.unit2unit(forces_units, "eV/Angstrom")
        self.stress_units = MDUnits.unit2unit(stress_units, "eV/A/A/A")

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """
        Args:
            atoms (ase.Atoms): ASE atoms object.
            properties (list of str): do not use this, no functionality
            system_changes (list of str): List of changes for ASE.
        """
        # First call original calculator to set atoms attribute
        # (see https://wiki.fysik.dtu.dk/ase/_modules/ase/calculators/calculator.html#Calculator)

        if self.calculation_required(atoms, properties):
            Calculator.calculate(self, atoms)
            # Convert to schnetpack input format
            model_inputs = self.atoms_converter(atoms)
            # Call model
            model_results = self.model(model_inputs)

            results = {}
            # Convert outputs to calculator format
            if self.model_energy is not None:
                if self.model_energy not in model_results.keys():
                    raise SpkCalculatorError(
                        "'{}' is not a property of your model. Please "
                        "check the model "
                        "properties!".format(self.model_energy)
                    )
                energy = model_results[self.model_energy].cpu().data.numpy()
                results[self.energy] = (
                    energy.item() * self.energy_units
                )  # ase calculator should return scalar energy

            if self.model_forces is not None:
                if self.model_forces not in model_results.keys():
                    raise SpkCalculatorError(
                        "'{}' is not a property of your model. Please "
                        "check the model"
                        "properties!".format(self.model_forces)
                    )
                forces = model_results[self.model_forces].cpu().data.numpy()
                results[self.forces] = (
                    forces.reshape((len(atoms), 3)) * self.forces_units
                )

            if self.model_stress is not None:
                if atoms.cell.volume <= 0.0:
                    raise SpkCalculatorError(
                        "Cell with 0 volume encountered for stress computation"
                    )

                if self.model_stress not in model_results.keys():
                    raise SpkCalculatorError(
                        "'{}' is not a property of your model. Please "
                        "check the model"
                        "properties! If desired, stress tensor computation can be "
                        "activated via schnetpack.utils.activate_stress_computation "
                        "at ones own risk.".format(self.model_stress)
                    )
                stress = model_results[self.model_stress].cpu().data.numpy()
                results[self.stress] = stress.reshape((3, 3)) * self.stress_units

            self.results = results


class AseInterface:
    """
    Interface for ASE calculations (optimization and molecular dynamics)

    Args:
        molecule_path (str): Path to initial geometry
        ml_model (object): Trained model
        working_dir (str): Path to directory where files should be stored
        device (str): cpu or cuda
    """

    def __init__(
        self,
        molecule_path,
        ml_model,
        working_dir,
        device="cpu",
        energy="energy",
        forces="forces",
        energy_units="eV",
        forces_units="eV/Angstrom",
        environment_provider=SimpleEnvironmentProvider(),
    ):
        # Setup directory
        self.working_dir = working_dir
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)

        # Load the molecule
        self.molecule = None
        self._load_molecule(molecule_path)

        # Set up calculator
        calculator = SpkCalculator(
            ml_model,
            device=device,
            energy=energy,
            forces=forces,
            energy_units=energy_units,
            forces_units=forces_units,
            environment_provider=environment_provider,
        )
        self.molecule.set_calculator(calculator)

        # Unless initialized, set dynamics to False
        self.dynamics = False

    def _load_molecule(self, molecule_path):
        """
        Load molecule from file (can handle all ase formats).

        Args:
            molecule_path (str): Path to molecular geometry
        """
        file_format = os.path.splitext(molecule_path)[-1]
        if file_format == "xyz":
            self.molecule = read_xyz(molecule_path)
        else:
            self.molecule = read(molecule_path)

    def save_molecule(self, name, file_format="xyz", append=False):
        """
        Save the current molecular geometry.

        Args:
            name (str): Name of save-file.
            file_format (str): Format to store geometry (default xyz).
            append (bool): If set to true, geometry is added to end of file
                (default False).
        """
        molecule_path = os.path.join(self.working_dir, "%s.%s" % (name, file_format))
        write(molecule_path, self.molecule, format=file_format, append=append)

    def calculate_single_point(self):
        """
        Perform a single point computation of the energies and forces and
        store them to the working directory. The format used is the extended
        xyz format. This functionality is mainly intended to be used for
        interfaces.
        """
        energy = self.molecule.get_potential_energy()
        forces = self.molecule.get_forces()
        self.molecule.energy = energy
        self.molecule.forces = forces

        self.save_molecule("single_point", file_format="extxyz")

    def init_md(
        self,
        name,
        time_step=0.5,
        temp_init=300,
        temp_bath=None,
        reset=False,
        interval=1,
    ):
        """
        Initialize an ase molecular dynamics trajectory. The logfile needs to
        be specifies, so that old trajectories are not overwritten. This
        functionality can be used to subsequently carry out equilibration and
        production.

        Args:
            name (str): Basic name of logfile and trajectory
            time_step (float): Time step in fs (default=0.5)
            temp_init (float): Initial temperature of the system in K
                (default is 300)
            temp_bath (float): Carry out Langevin NVT dynamics at the specified
                temperature. If set to None, NVE dynamics are performed
                instead (default=None)
            reset (bool): Whether dynamics should be restarted with new initial
                conditions (default=False)
            interval (int): Data is stored every interval steps (default=1)
        """

        # If a previous dynamics run has been performed, don't reinitialize
        # velocities unless explicitly requested via restart=True
        if not self.dynamics or reset:
            self._init_velocities(temp_init=temp_init)

        # Set up dynamics
        if temp_bath is None:
            self.dynamics = VelocityVerlet(self.molecule, time_step * units.fs)
        else:
            self.dynamics = Langevin(
                self.molecule,
                time_step * units.fs,
                temp_bath * units.kB,
                1.0 / (100.0 * units.fs),
            )

        # Create monitors for logfile and a trajectory file
        logfile = os.path.join(self.working_dir, "%s.log" % name)
        trajfile = os.path.join(self.working_dir, "%s.traj" % name)
        logger = MDLogger(
            self.dynamics,
            self.molecule,
            logfile,
            stress=False,
            peratom=False,
            header=True,
            mode="a",
        )
        trajectory = Trajectory(trajfile, "w", self.molecule)

        # Attach monitors to trajectory
        self.dynamics.attach(logger, interval=interval)
        self.dynamics.attach(trajectory.write, interval=interval)

    def _init_velocities(
        self, temp_init=300, remove_translation=True, remove_rotation=True
    ):
        """
        Initialize velocities for molecular dynamics

        Args:
            temp_init (float): Initial temperature in Kelvin (default 300)
            remove_translation (bool): Remove translation components of
                velocity (default True)
            remove_rotation (bool): Remove rotation components of velocity
                (default True)
        """
        MaxwellBoltzmannDistribution(self.molecule, temp_init * units.kB)
        if remove_translation:
            Stationary(self.molecule)
        if remove_rotation:
            ZeroRotation(self.molecule)

    def run_md(self, steps):
        """
        Perform a molecular dynamics simulation using the settings specified
        upon initializing the class.

        Args:
            steps (int): Number of simulation steps performed
        """
        if not self.dynamics:
            raise AttributeError(
                "Dynamics need to be initialized using the" " 'setup_md' function"
            )

        self.dynamics.run(steps)

    def optimize(self, fmax=1.0e-2, steps=1000):
        """
        Optimize a molecular geometry using the Quasi Newton optimizer in ase
        (BFGS + line search)

        Args:
            fmax (float): Maximum residual force change (default 1.e-2)
            steps (int): Maximum number of steps (default 1000)
        """
        name = "optimization"
        optimize_file = os.path.join(self.working_dir, name)
        optimizer = QuasiNewton(
            self.molecule,
            trajectory="%s.traj" % optimize_file,
            restart="%s.pkl" % optimize_file,
        )
        optimizer.run(fmax, steps)

        # Save final geometry in xyz format
        self.save_molecule(name)

    def compute_normal_modes(self, write_jmol=True):
        """
        Use ase calculator to compute numerical frequencies for the molecule

        Args:
            write_jmol (bool): Write frequencies to input file for
                visualization in jmol (default=True)
        """
        freq_file = os.path.join(self.working_dir, "normal_modes")

        # Compute frequencies
        frequencies = Vibrations(self.molecule, name=freq_file)
        frequencies.run()

        # Print a summary
        frequencies.summary()

        # Write jmol file if requested
        if write_jmol:
            frequencies.write_jmol()


MLPotential = DeprecationHelper(SpkCalculator, "MLPotential")



import argparse
import logging
import os

import schnetpack.datasets as dset

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))

if __name__ == "__main__":
    main_parser = argparse.ArgumentParser()
    data_subparsers = main_parser.add_subparsers(dest="dataset", help="Select dataset")
    data_subparsers.required = True
    qm9_parser = data_subparsers.add_parser("qm9", help="QM9 dataset")
    qm9_parser.add_argument("dbpath", help="Destination path")
    qm9_parser.add_argument(
        "--remove_uncharacterized",
        help="Remove uncharacterized molecules from QM9",
        type=bool,
        default="False",
    )

    md17_parser = data_subparsers.add_parser("md17", help="MD17 datasets")
    md17_parser.add_argument(
        "molecule", help="Molecule dataset", choices=dset.MD17.existing_datasets
    )
    md17_parser.add_argument("dbpath", help="Destination path")

    args = main_parser.parse_args()

    if args.dataset == "qm9":
        dset.QM9(args.dbpath, True)
    if args.dataset == "md17":
        dset.MD17(args.dbpath, args.molecule)
    else:
        print("Unknown dataset!")



"""
All molecular dynamics in SchNetPack is performed using the :obj:`schnetpack.md.Simulator` class.
This class collects the atomistic system (:obj:`schnetpack.md.System`), calculators (:obj:`schnetpack.md.calculators`),
integrators (:obj:`schnetpack.md.integrators`) and various simulation hooks (:obj:`schnetpack.md.simulation_hooks`)
and performs the time integration.
"""
from tqdm import tqdm


class Simulator:
    """
    Main driver of the molecular dynamics simulation. Uses an integrator to
    propagate the molecular system defined in the system class according to
    the forces yielded by a provided calculator.

    In addition, hooks can be applied at five different stages of each
    simulation step:
     - Start of the simulation (e.g. for initializing thermostats)
     - Before first integrator half step (e.g. thermostats)
     - After computation of the forces and before main integrator step (e.g.
      for accelerated MD)
     - After second integrator half step (e.g. thermostats, output routines)
     - At the end of the simulation (e.g. general wrap up of file writes, etc.)

    This routine has a state dict which can be used to restart a previous
    simulation.

    Args:
        system (schnetpack.md.System): Instance of the system class defined in
                         molecular_dynamics.system holding the structures,
                         masses, atom type, momenta, forces and properties of
                         all molecules and their replicas
        integrator (schnetpack.md.Integrator): Integrator for propagating the molecular
                             dynamics simulation, defined in
                             schnetpack.md.integrators
        calculator (schnetpack.md.calculator): Calculator class used to compute molecular
                             forces for propagation and (if requested)
                             various other properties.
        simulator_hooks (list(object)): List of different hooks to be applied
                                        during simulations. Examples would be
                                        file loggers and thermostats.
        step (int): Index of the initial simulation step.
        restart (bool): Indicates, whether the simulation is restarted. E.g. if set to True, the simulator tries to
                        continue logging in the previously created dataset. (default=False)
                        This is set automatically by the restart_simulation function. Enabling it without the function
                        currently only makes sense if independent simulations should be written to the same file.
    """

    def __init__(
        self, system, integrator, calculator, simulator_hooks=[], step=0, restart=False
    ):

        self.system = system
        self.integrator = integrator
        self.calculator = calculator
        self.simulator_hooks = simulator_hooks
        self.step = step
        self.n_steps = None
        self.restart = restart

        # Keep track of the actual simulation steps performed with simulate calls
        self.effective_steps = 0

    def simulate(self, n_steps=10000):
        """
        Main simulation function. Propagates the system for a certain number
        of steps.

        Args:
            n_steps (int): Number of simulation steps to be performed (
                           default=10000)
        """

        self.n_steps = n_steps

        # Perform initial computation of forces
        if self.system.forces is None:
            self.calculator.calculate(self.system)

        # Call hooks at the simulation start
        for hook in self.simulator_hooks:
            hook.on_simulation_start(self)

        for _ in tqdm(range(n_steps), ncols=120):

            # Call hook berfore first half step
            for hook in self.simulator_hooks:
                hook.on_step_begin(self)

            # Do half step momenta
            self.integrator.half_step(self.system)

            # Do propagation MD/PIMD
            self.integrator.main_step(self.system)

            # Compute new forces
            self.calculator.calculate(self.system)

            # Call hook after forces
            for hook in self.simulator_hooks:
                hook.on_step_middle(self)

            # Do half step momenta
            self.integrator.half_step(self.system)

            # Call hooks after second half step
            # Hooks are called in reverse order to guarantee symmetry of
            # the propagator when using thermostats and barostats
            for hook in self.simulator_hooks[::-1]:
                hook.on_step_end(self)

            self.step += 1
            self.effective_steps += 1

        # Call hooks at the simulation end
        for hook in self.simulator_hooks:
            hook.on_simulation_end(self)

    @property
    def state_dict(self):
        """
        State dict used to restart the simulation. Generates a dictionary with
        the following entries:
            - step: current simulation step
            - systems: state dict of the system holding current positions,
                       momenta, forces, etc...
            - simulator_hooks: dict of state dicts of the various hooks used
                               during simulation using their basic class
                               name as keys.

        Returns:
            dict: State dict containing the current step, the system
                  parameters (positions, momenta, etc.) and all
                  simulator_hook state dicts

        """
        state_dict = {
            "step": self.step,
            "system": self.system.state_dict,
            "simulator_hooks": {
                hook.__class__: hook.state_dict for hook in self.simulator_hooks
            },
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        """
        Set the current state dict of the simulator using a state dict
        defined in state_dict. This routine assumes, that the identity of all
        hooks has not changed and the order is preserved. A more general
        method to restart simulations is provided below.

        Args:
            state_dict (dict): state dict containing the entries 'step',
            'simulator_hooks' and 'system'.

        """
        self.step = state_dict["step"]
        self.system.state_dict = state_dict["system"]

        # Set state dicts of all hooks
        for hook in self.simulator_hooks:
            if hook.__class__ in state_dict["simulator_hooks"]:
                hook.state_dict = state_dict["simulator_hooks"][hook.__class__]

    def restart_simulation(self, state_dict, soft=False):
        """
        Routine for restarting a simulation. Reads the current step, as well
        as system state from the provided state dict. In case of the
        simulation hooks, only the states of the thermostat hooks are
        restored, as all other hooks do not depend on previous simulations.

        If the soft option is chosen, only restores states of thermostats if
        they are present in the current simulation and the state dict.
        Otherwise, all thermostats found in the state dict are required to be
        present in the current simulation.

        Args:
            state_dict (dict): State dict of the current simulation
            soft (bool): Flag to toggle hard/soft thermostat restarts (
                         default=False)

        """
        self.step = state_dict["step"]
        self.system.state_dict = state_dict["system"]

        if soft:
            # Do the same as in a basic state dict setting
            for hook in self.simulator_hooks:
                if hook.__class__ in state_dict["simulator_hooks"]:
                    hook.state_dict = state_dict["simulator_hooks"][hook.__class__]
        else:
            # Hard restart, require all thermostats to be there
            for hook in self.simulator_hooks:
                # Check if hook is thermostat
                if hasattr(hook, "temperature_bath"):
                    if hook.__class__ not in state_dict["simulator_hooks"]:
                        raise ValueError(
                            f"Could not find restart information for {hook.__class__} in state dict."
                        )
                    else:
                        hook.state_dict = state_dict["simulator_hooks"][hook.__class__]

        # In this case, set restart flag automatically
        self.restart = True

    def load_system_state(self, state_dict):
        """
        Routine for only loading the system state of previous simulations.
        This can e.g. be used for production runs, where an equilibrated system
        is loaded, but the thermostat is changed.

        Args:
            state_dict (dict): State dict of the current simulation
        """
        self.system.state_dict = state_dict["system"]



import numpy as np
import torch
from torch import nn as nn
from torch.autograd import grad

import schnetpack
from schnetpack import nn as L, Properties

__all__ = [
    "Atomwise",
    "ElementalAtomwise",
    "DipoleMoment",
    "ElementalDipoleMoment",
    "Polarizability",
    "ElectronicSpatialExtent",
]


class AtomwiseError(Exception):
    pass


class Atomwise(nn.Module):
    """
    Predicts atom-wise contributions and accumulates global prediction, e.g. for the
    energy.

    Args:
        n_in (int): input dimension of representation
        n_out (int): output dimension of target property (default: 1)
        aggregation_mode (str): one of {sum, avg} (default: sum)
        n_layers (int): number of nn in output network (default: 2)
        n_neurons (list of int or None): number of neurons in each layer of the output
            network. If `None`, divide neurons by 2 in each layer. (default: None)
        activation (function): activation function for hidden nn
            (default: spk.nn.activations.shifted_softplus)
        property (str): name of the output property (default: "y")
        contributions (str or None): Name of property contributions in return dict.
            No contributions returned if None. (default: None)
        derivative (str or None): Name of property derivative. No derivative
            returned if None. (default: None)
        negative_dr (bool): Multiply the derivative with -1 if True. (default: False)
        stress (str or None): Name of stress property. Compute the derivative with
            respect to the cell parameters if not None. (default: None)
        create_graph (bool): If False, the graph used to compute the grad will be
            freed. Note that in nearly all cases setting this option to True is not
            needed and often can be worked around in a much more efficient way.
            Defaults to the value of create_graph. (default: False)
        mean (torch.Tensor or None): mean of property
        stddev (torch.Tensor or None): standard deviation of property (default: None)
        atomref (torch.Tensor or None): reference single-atom properties. Expects
            an (max_z + 1) x 1 array where atomref[Z] corresponds to the reference
            property of element Z. The value of atomref[0] must be zero, as this
            corresponds to the reference property for for "mask" atoms. (default: None)
        outnet (callable): Network used for atomistic outputs. Takes schnetpack input
            dictionary as input. Output is not normalized. If set to None,
            a pyramidal network is generated automatically. (default: None)

    Returns:
        tuple: prediction for property

        If contributions is not None additionally returns atom-wise contributions.

        If derivative is not None additionally returns derivative w.r.t. atom positions.

    """

    def __init__(
        self,
        n_in,
        n_out=1,
        aggregation_mode="sum",
        n_layers=2,
        n_neurons=None,
        activation=schnetpack.nn.activations.shifted_softplus,
        property="y",
        contributions=None,
        derivative=None,
        negative_dr=False,
        stress=None,
        create_graph=False,
        mean=None,
        stddev=None,
        atomref=None,
        outnet=None,
    ):
        super(Atomwise, self).__init__()

        self.n_layers = n_layers
        self.create_graph = create_graph
        self.property = property
        self.contributions = contributions
        self.derivative = derivative
        self.negative_dr = negative_dr
        self.stress = stress

        mean = torch.FloatTensor([0.0]) if mean is None else mean
        stddev = torch.FloatTensor([1.0]) if stddev is None else stddev

        # initialize single atom energies
        if atomref is not None:
            self.atomref = nn.Embedding.from_pretrained(
                torch.from_numpy(atomref.astype(np.float32))
            )
        else:
            self.atomref = None

        # build output network
        if outnet is None:
            self.out_net = nn.Sequential(
                schnetpack.nn.base.GetItem("representation"),
                schnetpack.nn.blocks.MLP(n_in, n_out, n_neurons, n_layers, activation),
            )
        else:
            self.out_net = outnet

        # build standardization layer
        self.standardize = schnetpack.nn.base.ScaleShift(mean, stddev)

        # build aggregation layer
        if aggregation_mode == "sum":
            self.atom_pool = schnetpack.nn.base.Aggregate(axis=1, mean=False)
        elif aggregation_mode == "avg":
            self.atom_pool = schnetpack.nn.base.Aggregate(axis=1, mean=True)
        elif aggregation_mode == "max":
            self.atom_pool = schnetpack.nn.base.MaxAggregate(axis=1)
        elif aggregation_mode == "softmax":
            self.atom_pool = schnetpack.nn.base.SoftmaxAggregate(axis=1)
        else:
            raise AtomwiseError(
                "{} is not a valid aggregation " "mode!".format(aggregation_mode)
            )

    def forward(self, inputs):
        r"""
        predicts atomwise property
        """
        atomic_numbers = inputs[Properties.Z]
        atom_mask = inputs[Properties.atom_mask]

        # run prediction
        yi = self.out_net(inputs)
        yi = self.standardize(yi)

        if self.atomref is not None:
            y0 = self.atomref(atomic_numbers)
            yi = yi + y0

        y = self.atom_pool(yi, atom_mask)

        # collect results
        result = {self.property: y}

        if self.contributions is not None:
            result[self.contributions] = yi

        create_graph = True if self.training else self.create_graph

        if self.derivative is not None:
            sign = -1.0 if self.negative_dr else 1.0
            dy = grad(
                result[self.property],
                inputs[Properties.R],
                grad_outputs=torch.ones_like(result[self.property]),
                create_graph=create_graph,
                retain_graph=True,
            )[0]
            result[self.derivative] = sign * dy

        if self.stress is not None:
            cell = inputs[Properties.cell]
            # Compute derivative with respect to cell displacements
            stress = grad(
                result[self.property],
                inputs["displacement"],
                grad_outputs=torch.ones_like(result[self.property]),
                create_graph=create_graph,
                retain_graph=True,
            )[0]
            # Compute cell volume
            volume = torch.sum(
                cell[:, 0, :] * torch.cross(cell[:, 1, :], cell[:, 2, :], dim=1),
                dim=1,
                keepdim=True,
            )[..., None]
            # Finalize stress tensor
            result[self.stress] = stress / volume

        return result


class DipoleMoment(Atomwise):
    """
    Predicts latent partial charges and calculates dipole moment.

    Args:
        n_in (int): input dimension of representation
        n_layers (int): number of layers in output network (default: 2)
        n_neurons (list of int or None): number of neurons in each layer of the output
            network. If `None`, divide neurons by 2 in each layer. (default: None)
        activation (torch.Function): activation function for hidden nn
            (default: schnetpack.nn.activations.shifted_softplus)
        property (str): name of the output property (default: "y")
        contributions (str or None): Name of property contributions in return dict.
            No contributions returned if None. (default: None)
        charge_correction (str or None): Name of charge labels in dataset. If
            something is selected, the charge contributions are corrected according
            to the total charges in the dataset. No charge correction if None.
            (default: None)
        predict_magnitude (bool): if True, predict the magnitude of the dipole moment
            instead of the vector (default: False)
        mean (torch.FloatTensor or None): mean of dipole (default: None)
        stddev (torch.FloatTensor or None): stddev of dipole (default: None)

    Returns:
        dict: vector for the dipole moment

        If predict_magnitude is True returns the magnitude of the dipole moment
        instead of the vector.

        If contributions is not None latent atomic charges are added to the output
        dictionary.
    """

    def __init__(
        self,
        n_in,
        n_layers=2,
        n_neurons=None,
        activation=schnetpack.nn.activations.shifted_softplus,
        property="y",
        contributions=None,
        charge_correction=None,
        predict_magnitude=False,
        mean=None,
        stddev=None,
        outnet=None,
    ):
        self.predict_magnitude = predict_magnitude
        self.charge_correction = charge_correction
        super(DipoleMoment, self).__init__(
            n_in,
            1,
            "sum",
            n_layers,
            n_neurons,
            activation=activation,
            mean=mean,
            stddev=stddev,
            outnet=outnet,
            property=property,
            contributions=contributions,
        )

    def forward(self, inputs):
        """
        predicts dipole moment
        """
        positions = inputs[Properties.R]
        atom_mask = inputs[Properties.atom_mask]

        # run prediction
        charges = self.out_net(inputs) * atom_mask[:, :, None]

        # charge correction
        if self.charge_correction is not None:
            total_charges = inputs[self.charge_correction]
            charge_correction = total_charges - charges.sum(1)
            charges = charges + (
                charge_correction / atom_mask.sum(-1).unsqueeze(-1)
            ).unsqueeze(-1)
            charges *= atom_mask[:, :, None]

        yi = positions * charges
        y = self.atom_pool(yi)

        if self.predict_magnitude:
            y = torch.norm(y, dim=1, keepdim=True)

        # collect results
        result = {self.property: y}

        if self.contributions:
            result[self.contributions] = charges

        return result


class ElementalAtomwise(Atomwise):
    """
    Predicts properties in atom-wise fashion using a separate network for every chemical
    element of the central atom. Particularly useful for networks of
    Behler-Parrinello type.

    Args:
        n_in (int): input dimension of representation (default: 128)
        n_out (int): output dimension of target property (default: 1)
        aggregation_mode (str): one of {sum, avg} (default: sum)
        n_layers (int): number of nn in output network (default: 3)
        property (str): name of the output property (default: "y")
        derivative (str or None): Name of property derivative. No derivative
            returned if None. (default: None)
        negative_dr (bool): Multiply the derivative with -1 if True. (default: False)
        contributions (str or None): Name of property contributions in return dict.
            No contributions returned if None. (default: None)
        create_graph (bool): If False, the graph used to compute the grad will be
            freed. Note that in nearly all cases setting this option to True is not nee
            ded and often can be worked around in a much more efficient way. Defaults to
            the value of create_graph. (default: False)
        elements (set of int): List of atomic numbers present in the training set
            {1,6,7,8,9} for QM9. (default: frozenset(1,6,7,8,9))
        n_hidden (int): number of neurons in each layer of the output network.
            (default: 50)
        activation (function): activation function for hidden nn
            (default: spk.nn.activations.shifted_softplus)
        mean (torch.Tensor or None): mean of property
        stddev (torch.Tensor or None): standard deviation of property (default: None)
        atomref (torch.Tensor or None): reference single-atom properties. Expects
            an (max_z + 1) x 1 array where atomref[Z] corresponds to the reference
            property of element Z. The value of atomref[0] must be zero, as this
            corresponds to the reference property for for "mask" atoms. (default: None)
    """

    def __init__(
        self,
        n_in=128,
        n_out=1,
        aggregation_mode="sum",
        n_layers=3,
        property="y",
        derivative=None,
        negative_dr=False,
        contributions=None,
        create_graph=True,
        elements=frozenset((1, 6, 7, 8, 9)),
        n_hidden=50,
        activation=schnetpack.nn.activations.shifted_softplus,
        mean=None,
        stddev=None,
        atomref=None,
    ):
        outnet = schnetpack.nn.blocks.GatedNetwork(
            n_in,
            n_out,
            elements,
            n_hidden=n_hidden,
            n_layers=n_layers,
            activation=activation,
        )

        super(ElementalAtomwise, self).__init__(
            n_in=n_in,
            n_out=n_out,
            aggregation_mode=aggregation_mode,
            n_layers=n_layers,
            n_neurons=None,
            activation=activation,
            property=property,
            contributions=contributions,
            derivative=derivative,
            negative_dr=negative_dr,
            create_graph=create_graph,
            mean=mean,
            stddev=stddev,
            atomref=atomref,
            outnet=outnet,
        )


class ElementalDipoleMoment(DipoleMoment):
    """
    Predicts partial charges and computes dipole moment using serparate NNs for every different element.
    Particularly useful for networks of Behler-Parrinello type.

    Args:
        n_in (int): input dimension of representation
        n_out (int): output dimension of representation (default: 1)
        n_layers (int): number of layers in output network (default: 3)
        predict_magnitude (bool): if True, predict the magnitude of the dipole moment instead of the vector (default: False)
        elements (set of int): List of atomic numbers present in the training set {1,6,7,8,9} for QM9. (default: frozenset(1,6,7,8,9))
        n_hidden (int): number of neurons in each layer of the output network. (default: 50)
        activation (function): activation function for hidden nn (default: schnetpack.nn.activations.shifted_softplus)
        activation (function): activation function for hidden nn
        mean (torch.FloatTensor): mean of energy
        stddev (torch.FloatTensor): standard deviation of energy
    """

    def __init__(
        self,
        n_in,
        n_out=1,
        n_layers=3,
        contributions=False,
        property="y",
        predict_magnitude=False,
        elements=frozenset((1, 6, 7, 8, 9)),
        n_hidden=50,
        activation=schnetpack.nn.activations.shifted_softplus,
        mean=None,
        stddev=None,
    ):
        outnet = schnetpack.nn.blocks.GatedNetwork(
            n_in,
            n_out,
            elements,
            n_hidden=n_hidden,
            n_layers=n_layers,
            activation=activation,
        )

        super(ElementalDipoleMoment, self).__init__(
            n_in,
            n_layers,
            None,
            activation=activation,
            property=property,
            contributions=contributions,
            outnet=outnet,
            predict_magnitude=predict_magnitude,
            mean=mean,
            stddev=stddev,
        )


class Polarizability(Atomwise):
    """
    Predicts polarizability of input molecules.

    Args:
        n_in (int): input dimension of representation (default: 128)
        aggregation_mode (str): one of {sum, avg} (default: sum)
        n_layers (int): number of nn in output network (default: 2)
        n_neurons (list of int or None): number of neurons in each layer of the output
            network. If `None`, divide neurons by 2 in each layer. (default: None)
        activation (function): activation function for hidden nn
            (default: spk.nn.activations.shifted_softplus)
        property (str): name of the output property (default: "y")
        isotropic (bool): return isotropic polarizability if True. (default: False)
        create_graph (bool): If False, the graph used to compute the grad will be
            freed. Note that in nearly all cases setting this option to True is not nee
            ded and often can be worked around in a much more efficient way. Defaults to
            the value of create_graph. (default: False)
        outnet (callable): Network used for atomistic outputs. Takes schnetpack input
            dictionary as input. Output is not normalized. If set to None,
            a pyramidal network is generated automatically. (default: None)
        cutoff_network (nn.Module): cutoff network (default: None)

    Returns:
        dict: Polarizability of molecules

        Adds isotropic polarizability if isotropic is not None.
    """

    def __init__(
        self,
        n_in=128,
        aggregation_mode="sum",
        n_layers=2,
        n_neurons=None,
        activation=L.shifted_softplus,
        property="y",
        isotropic=False,
        create_graph=True,
        outnet=None,
        cutoff_network=None,
    ):
        super(Polarizability, self).__init__(
            n_in=n_in,
            n_out=2,
            n_layers=n_layers,
            aggregation_mode=aggregation_mode,
            n_neurons=n_neurons,
            activation=activation,
            property=property,
            derivative=None,
            create_graph=create_graph,
            outnet=outnet,
        )
        self.isotropic = isotropic
        self.nbh_agg = L.Aggregate(2)
        self.atom_agg = L.Aggregate(1)

        self.cutoff_network = cutoff_network

    def forward(self, inputs):
        positions = inputs[Properties.R]
        neighbors = inputs[Properties.neighbors]
        nbh_mask = inputs[Properties.neighbor_mask]
        atom_mask = inputs[Properties.atom_mask]

        # Get environment distances and positions
        distances, dist_vecs = L.atom_distances(positions, neighbors, return_vecs=True)

        # Get atomic contributions
        contributions = self.out_net(inputs)

        # Redistribute contributions to neighbors
        # B x A x N x 1
        # neighbor_contributions = L.neighbor_elements(c1, neighbors)
        neighbor_contributions = L.neighbor_elements(contributions, neighbors)

        if self.cutoff_network is not None:
            f_cut = self.cutoff_network(distances)[..., None]
            neighbor_contributions = neighbor_contributions * f_cut

        neighbor_contributions1 = neighbor_contributions[:, :, :, 0]
        neighbor_contributions2 = neighbor_contributions[:, :, :, 1]

        # B x A x N x 3
        atomic_dipoles = self.nbh_agg(
            dist_vecs * neighbor_contributions1[..., None], nbh_mask
        )
        # B x A x N x 3

        masked_dist = (distances ** 3 * nbh_mask) + (1 - nbh_mask)
        nbh_fields = (
            dist_vecs * neighbor_contributions2[..., None] / masked_dist[..., None]
        )
        atomic_fields = self.nbh_agg(nbh_fields, nbh_mask)
        field_norm = torch.norm(atomic_fields, dim=-1, keepdim=True)
        field_norm = field_norm + (field_norm < 1e-10).float()
        atomic_fields = atomic_fields / field_norm

        atomic_polar = atomic_dipoles[..., None] * atomic_fields[:, :, None, :]

        # Symmetrize
        atomic_polar = symmetric_product(atomic_polar)

        global_polar = self.atom_agg(atomic_polar, atom_mask[..., None])

        result = {
            # "y_i": atomic_polar,
            self.property: global_polar
        }

        if self.isotropic:
            result[self.property] = torch.mean(
                torch.diagonal(global_polar, dim1=-2, dim2=-1), dim=-1, keepdim=True
            )
        return result


def symmetric_product(tensor):
    """
    Symmetric outer product of tensor
    """
    shape = tensor.size()
    idx = list(range(len(shape)))
    idx[-1], idx[-2] = idx[-2], idx[-1]
    return 0.5 * (tensor + tensor.permute(*idx))


class ElectronicSpatialExtent(Atomwise):
    """
    Predicts the electronic spatial extent using a formalism close to the dipole moment layer.
    The electronic spatial extent is discretized as a sum of atomic latent contributions
    weighted by the squared distance of the atom from the center of mass (SchNetPack default).

    .. math:: ESE = \sum_i^N | R_{i0} |^2 q(R)

    Args:
        n_in (int): input dimension of representation
        n_layers (int): number of layers in output network (default: 2)
        n_neurons (list of int or None): number of neurons in each layer of the output
            network. If `None`, divide neurons by 2 in each layer. (default: None)
        activation (torch.Function): activation function for hidden nn
            (default: schnetpack.nn.activations.shifted_softplus)
        property (str): name of the output property (default: "y")
        contributions (str or None): Name of property contributions in return dict.
            No contributions returned if None. (default: None)
        mean (torch.FloatTensor or None): mean of dipole (default: None)
        stddev (torch.FloatTensor or None): stddev of dipole (default: None)

    Returns:
        dict: the electronic spatial extent

        If contributions is not None latent atomic charges are added to the output
        dictionary.
    """

    def __init__(
        self,
        n_in,
        n_layers=2,
        n_neurons=None,
        activation=schnetpack.nn.activations.shifted_softplus,
        property="y",
        contributions=None,
        mean=None,
        stddev=None,
        outnet=None,
    ):
        super(ElectronicSpatialExtent, self).__init__(
            n_in,
            1,
            "sum",
            n_layers,
            n_neurons,
            activation=activation,
            mean=mean,
            stddev=stddev,
            outnet=outnet,
            property=property,
            contributions=contributions,
        )

    def forward(self, inputs):
        """
        Predicts the electronic spatial extent.
        """
        positions = inputs[Properties.R]
        atom_mask = inputs[Properties.atom_mask][:, :, None]

        # run prediction
        charges = self.out_net(inputs) * atom_mask
        yi = torch.norm(positions, 2, 2, keepdim=True) ** 2 * charges
        y = self.atom_pool(yi)

        # collect results
        result = {self.property: y}

        if self.contributions:
            result[self.contributions] = charges

        return result



import schnetpack as spk
from schnetpack.datasets import (
    QM9,
    ANI1,
    MD17,
    MaterialsProject,
    OrganicMaterialsDatabase,
)


__all__ = [
    "divide_by_atoms",
    "pooling_mode",
    "get_divide_by_atoms",
    "get_pooling_mode",
    "get_negative_dr",
    "get_derivative",
    "get_contributions",
    "get_stress",
    "get_module_str",
    "get_environment_provider",
]


divide_by_atoms = {
    QM9.mu: True,
    QM9.alpha: True,
    QM9.homo: False,
    QM9.lumo: False,
    QM9.gap: False,
    QM9.r2: True,
    QM9.zpve: True,
    QM9.U0: True,
    QM9.U: True,
    QM9.H: True,
    QM9.G: True,
    QM9.Cv: True,
    ANI1.energy: True,
    MD17.energy: True,
    MaterialsProject.EformationPerAtom: False,
    MaterialsProject.EPerAtom: False,
    MaterialsProject.BandGap: False,
    MaterialsProject.TotalMagnetization: True,
    OrganicMaterialsDatabase.BandGap: False,
}

pooling_mode = {
    QM9.mu: "sum",
    QM9.alpha: "sum",
    QM9.homo: "avg",
    QM9.lumo: "avg",
    QM9.gap: "avg",
    QM9.r2: "sum",
    QM9.zpve: "sum",
    QM9.U0: "sum",
    QM9.U: "sum",
    QM9.H: "sum",
    QM9.G: "sum",
    QM9.Cv: "sum",
    ANI1.energy: "sum",
    MD17.energy: "sum",
    MaterialsProject.EformationPerAtom: "avg",
    MaterialsProject.EPerAtom: "avg",
    MaterialsProject.BandGap: "avg",
    MaterialsProject.TotalMagnetization: "sum",
    OrganicMaterialsDatabase.BandGap: "avg",
}


def get_divide_by_atoms(args):
    """
    Get 'divide_by_atoms'-parameter depending on run arguments.
    """
    if args.dataset == "custom":
        return args.aggregation_mode == "sum"
    return divide_by_atoms[args.property]


def get_pooling_mode(args):
    """
    Get 'pooling_mode'-parameter depending on run arguments.
    """
    if args.dataset == "custom":
        return args.aggregation_mode
    return pooling_mode[args.property]


def get_derivative(args):
    if args.dataset == "custom":
        return args.derivative
    elif args.dataset == "md17" and not args.ignore_forces:
        return spk.datasets.MD17.forces
    return None


def get_contributions(args):
    if args.dataset == "custom":
        return args.contributions
    return None


def get_stress(args):
    if args.dataset == "custom":
        return args.stress
    return None


def get_negative_dr(args):
    if args.dataset == "custom":
        return args.negative_dr
    elif args.dataset == "md17":
        return True
    return False


def get_module_str(args):
    if args.dataset == "custom":
        return args.output_module
    if args.model == "schnet":
        if args.property == spk.datasets.QM9.mu:
            return "dipole_moment"
        if args.property == spk.datasets.QM9.r2:
            return "electronic_spatial_extent"
        return "atomwise"
    elif args.model == "wacsf":
        if args.property == spk.datasets.QM9.mu:
            return "elemental_dipole_moment"
        return "elemental_atomwise"


def get_environment_provider(args, device):
    if args.environment_provider == "simple":
        return spk.environment.SimpleEnvironmentProvider()
    elif args.environment_provider == "ase":
        return spk.environment.AseEnvironmentProvider(cutoff=args.cutoff)
    elif args.environment_provider == "torch":
        return spk.environment.TorchEnvironmentProvider(
            cutoff=args.cutoff, device=device
        )
    else:
        raise NotImplementedError



import logging

import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader

logger = logging.getLogger(__name__)

from schnetpack import Properties
from .stats import StatisticsAccumulator

__all__ = ["AtomsLoader"]


def _collate_aseatoms(examples):
    """
    Build batch from systems and properties & apply padding

    Args:
        examples (list):

    Returns:
        dict[str->torch.Tensor]: mini-batch of atomistic systems
    """
    properties = examples[0]

    # initialize maximum sizes
    max_size = {
        prop: np.array(val.size(), dtype=np.int) for prop, val in properties.items()
    }

    # get maximum sizes
    for properties in examples[1:]:
        for prop, val in properties.items():
            max_size[prop] = np.maximum(
                max_size[prop], np.array(val.size(), dtype=np.int)
            )

    # initialize batch
    batch = {
        p: torch.zeros(len(examples), *[int(ss) for ss in size]).type(
            examples[0][p].type()
        )
        for p, size in max_size.items()
    }
    has_atom_mask = Properties.atom_mask in batch.keys()
    has_neighbor_mask = Properties.neighbor_mask in batch.keys()

    if not has_neighbor_mask:
        batch[Properties.neighbor_mask] = torch.zeros_like(
            batch[Properties.neighbors]
        ).float()
    if not has_atom_mask:
        batch[Properties.atom_mask] = torch.zeros_like(batch[Properties.Z]).float()

    # If neighbor pairs are requested, construct mask placeholders
    # Since the structure of both idx_j and idx_k is identical
    # (not the values), only one cutoff mask has to be generated
    if Properties.neighbor_pairs_j in properties:
        batch[Properties.neighbor_pairs_mask] = torch.zeros_like(
            batch[Properties.neighbor_pairs_j]
        ).float()

    # build batch and pad
    for k, properties in enumerate(examples):
        for prop, val in properties.items():
            shape = val.size()
            s = (k,) + tuple([slice(0, d) for d in shape])
            batch[prop][s] = val

        # add mask
        if not has_neighbor_mask:
            nbh = properties[Properties.neighbors]
            shape = nbh.size()
            s = (k,) + tuple([slice(0, d) for d in shape])
            mask = nbh >= 0
            batch[Properties.neighbor_mask][s] = mask
            batch[Properties.neighbors][s] = nbh * mask.long()

        if not has_atom_mask:
            z = properties[Properties.Z]
            shape = z.size()
            s = (k,) + tuple([slice(0, d) for d in shape])
            batch[Properties.atom_mask][s] = z > 0

        # Check if neighbor pair indices are present
        # Since the structure of both idx_j and idx_k is identical
        # (not the values), only one cutoff mask has to be generated
        if Properties.neighbor_pairs_j in properties:
            nbh_idx_j = properties[Properties.neighbor_pairs_j]
            shape = nbh_idx_j.size()
            s = (k,) + tuple([slice(0, d) for d in shape])
            batch[Properties.neighbor_pairs_mask][s] = nbh_idx_j >= 0

    return batch


class AtomsLoader(DataLoader):
    r"""
    Specialized ``torch.data.DataLoader`` which uses the correct
    collate_fn for AtomsData and provides functionality for calculating mean
    and stddev.

    Arguments:
        dataset (Dataset): dataset from which to load the data.
        batch_size (int, optional): how many samples per batch to load
            (default: 1).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: False).
        sampler (Sampler, optional): defines the strategy to draw samples from
            the dataset. If specified, ``shuffle`` must be False.
        batch_sampler (Sampler, optional): like sampler, but returns a batch of
            indices at a time. Mutually exclusive with batch_size, shuffle,
            sampler, and drop_last.
        num_workers (int, optional): how many subprocesses to use for data
            loading. 0 means that the data will be loaded in the main process.
            (default: 0)
        collate_fn (callable, optional): merges a list of samples to form a
            mini-batch (default: collate_atons).
        pin_memory (bool, optional): If ``True``, the data loader will copy
            tensors into CUDA pinned memory before returning them.
        drop_last (bool, optional): set to ``True`` to drop the last incomplete
            batch, if the dataset size is not divisible by the batch size.
            If ``False`` and the size of dataset is not divisible by the batch
            size, then the last batch will be smaller. (default: False)
        timeout (numeric, optional): if positive, the timeout value for
            collecting a batch from workers. Should always be non-negative.
            (default: 0)
        worker_init_fn (callable, optional): If not None, this will be called
            on each worker subprocess with the worker id (an int in
            ``[0, num_workers - 1]``) as input, after seeding and before data
            loading. (default: None)

    """

    def __init__(
        self,
        dataset,
        batch_size=1,
        shuffle=False,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        collate_fn=_collate_aseatoms,
        pin_memory=False,
        drop_last=False,
        timeout=0,
        worker_init_fn=None,
    ):
        super(AtomsLoader, self).__init__(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
        )

    def get_statistics(
        self, property_names, divide_by_atoms=False, single_atom_ref=None
    ):
        """
        Compute mean and variance of a property. Uses the incremental Welford
        algorithm implemented in StatisticsAccumulator

        Args:
            property_names (str or list): Name of the property for which the
                mean and standard deviation should be computed
            divide_by_atoms (dict or bool): divide mean by number of atoms if True
                (default: False)
            single_atom_ref (dict or bool): reference values for single atoms (default:
                None)

        Returns:
            mean: Mean value
            stddev: Standard deviation

        """
        if type(property_names) is not list:
            property_names = [property_names]
        if type(divide_by_atoms) is not dict:
            divide_by_atoms = {prop: divide_by_atoms for prop in property_names}
        if single_atom_ref is None:
            single_atom_ref = {prop: None for prop in property_names}

        with torch.no_grad():
            statistics = {
                prop: StatisticsAccumulator(batch=True) for prop in property_names
            }
            logger.info("statistics will be calculated...")

            for row in self:
                for prop in property_names:
                    self._update_statistic(
                        divide_by_atoms[prop],
                        single_atom_ref[prop],
                        prop,
                        row,
                        statistics[prop],
                    )

            means = {prop: s.get_mean() for prop, s in statistics.items()}
            stddevs = {prop: s.get_stddev() for prop, s in statistics.items()}

        return means, stddevs

    def _update_statistic(
        self, divide_by_atoms, single_atom_ref, property_name, row, statistics
    ):
        """
        Helper function to update iterative mean / stddev statistics
        """
        property_value = row[property_name]
        if single_atom_ref is not None:
            z = row["_atomic_numbers"]
            p0 = torch.sum(torch.from_numpy(single_atom_ref[z]).float(), dim=1)
            property_value -= p0
        if divide_by_atoms:
            mask = torch.sum(row["_atom_mask"], dim=1, keepdim=True).view(
                [-1, 1] + [1] * (property_value.dim() - 2)
            )
            property_value /= mask
        statistics.add_sample(property_value)



"""
This module is used to store all information on the simulated atomistic systems.
It includes functionality for loading molecules from files.
All this functionality is encoded in the :obj:`schnetpack.md.System` class.
"""
import torch
from ase.io import read

from schnetpack.md.utils import MDUnits, compute_centroid, batch_inverse
from ase import Atoms


class SystemException(Exception):
    pass


class System:
    """
    Container for all properties associated with the simulated molecular system
    (masses, positions, momenta, ...). Uses atomic units internally.

    In order to simulate multiple systems efficiently dynamics properties
    (positions, momenta, forces) are torch tensors with the following
    dimensions:
        n_replicas x n_molecules x n_atoms x 3

    Here n_replicas is the number of copies for every molecule. In a normal
    simulation, these are treated as independent molecules e.g. for sampling
    purposes. In the case of ring polymer molecular dynamics (using the
    RingPolymer integrator), these replicas correspond to the beads of the
    polymer. n_molecules is the number of different molecules constituting
    the system, these can e.g. be different initial configurations of the
    same system (once again for sampling) or completely different molecules. In
    the latter case, the maximum number of atoms n_atoms (3rd dimension) is
    determined automatically and all arrays padded with zeros.

    Static properties (n_atoms, masses, atom_types and atom_masks) are stored in
    tensors of the shape:
        n_atoms : 1 x n_molecules (the same for all replicas)
        masses : 1 x n_molecules x n_atoms x 1 (the same for all replicas)
        atom_types : n_replicas x n_molecules x n_atoms x 1 (are brought to this
                     shape in order to avoid reshapes during every calculator
                     call)
        atom_masks : n_replicas x n_molecules x n_atoms x 1 (can change if
                     neighbor lists change for the replicas)

    n_atoms contains the number of atoms present in every molecule, masses
    and atom_types contain the molcular masses and nuclear charges.
    atom_masks is a binary array used to mask superfluous entries introduced
    by the zero-padding for differently sized molecules.

    Finally a dictionary properties stores the results of every calculator
    call for easy access of e.g. energies and dipole moments.

    Args:
        n_replicas (int): Number of replicas generated for each molecule.
        device (str): Computation device (default='cuda').
    """

    def __init__(self, n_replicas, device="cuda", initializer=None):

        # Specify device
        self.device = device

        # number of molecules, replicas of each and vector with the number of
        # atoms in each molecule
        self.n_replicas = n_replicas
        self.n_molecules = None
        self.n_atoms = None
        self.max_n_atoms = None

        # General static molecular properties
        self.atom_types = None
        self.masses = None
        self.atom_masks = None

        # Dynamic properties updated during simulation
        self.positions = None
        self.momenta = None
        self.forces = None
        self.energies = None

        # Properties for periodic boundary conditions and crystal cells
        self.cells = None
        self.pbc = None
        self.stress = None  # Used for the computation of the pressure

        # Property dictionary, updated during simulation
        self.properties = {}

        # Initialize initial conditions
        self.initializer = initializer

    def load_molecules_from_xyz(self, path_to_file):
        """
        Wrapper for loading molecules from .xyz file

        Args:
            path_to_file (str): path to data-file

        """
        molecules = read(path_to_file)
        if not type(molecules) == list:
            molecules = [molecules]
        self.load_molecules(molecules=molecules)

    def load_molecules(self, molecules):
        """
        Initializes all required variables and tensors based on a list of ASE
        atoms objects.

        Args:
            molecules list(ase.Atoms): List of ASE atoms objects containing
            molecular structures and chemical elements.
        """
        # 0) Check if molecules is a single ase.Atoms object and wrap it in list.
        if isinstance(molecules, Atoms):
            molecules = [molecules]

        # 1) Get maximum number of molecules, number of replicas and number of
        #    overall systems
        self.n_molecules = len(molecules)

        # 2) Construct array with number of atoms in each molecule
        self.n_atoms = torch.zeros(
            self.n_molecules, dtype=torch.long, device=self.device
        )

        for i in range(self.n_molecules):
            self.n_atoms[i] = molecules[i].get_global_number_of_atoms()

        # 3) Determine the maximum number of atoms present (in case of
        #    differently sized molecules)
        self.max_n_atoms = int(torch.max(self.n_atoms))

        # 4) Construct basic properties and masks
        self.atom_types = torch.zeros(
            self.n_replicas, self.n_molecules, self.max_n_atoms, device=self.device
        ).long()
        self.atom_masks = torch.zeros(
            self.n_replicas, self.n_molecules, self.max_n_atoms, device=self.device
        )
        self.masses = torch.ones(self.n_molecules, self.max_n_atoms, device=self.device)

        # Relevant for dynamic properties: positions, momenta, forces
        self.positions = torch.zeros(
            self.n_replicas, self.n_molecules, self.max_n_atoms, 3, device=self.device
        )
        self.momenta = torch.zeros(
            self.n_replicas, self.n_molecules, self.max_n_atoms, 3, device=self.device
        )

        # Relevant for periodic boundary conditions and simulation cells
        self.cells = torch.zeros(
            self.n_replicas, self.n_molecules, 3, 3, device=self.device
        )
        self.pbc = torch.zeros(self.n_molecules, 3, device=self.device)

        # 5) Populate arrays according to the data provided in molecules
        for i in range(self.n_molecules):
            # Static properties
            self.atom_types[:, i, : self.n_atoms[i]] = torch.from_numpy(
                molecules[i].get_atomic_numbers()
            )
            self.atom_masks[:, i, : self.n_atoms[i]] = 1.0
            self.masses[i, : self.n_atoms[i]] = torch.from_numpy(
                molecules[i].get_masses() * MDUnits.da2internal
            )

            # Dynamic properties
            self.positions[:, i, : self.n_atoms[i], :] = torch.from_numpy(
                molecules[i].positions * MDUnits.angs2internal
            )
            # Properties for cell simulations
            self.cells[:, i, :, :] = torch.from_numpy(
                molecules[i].cell * MDUnits.angs2internal
            )
            self.pbc[i, :] = torch.from_numpy(molecules[i].pbc)

        # Convert periodic boundary conditions to Boolean tensor
        self.pbc = self.pbc.bool()

        # Check for cell/pbc stuff:
        if torch.sum(torch.abs(self.cells)) == 0.0:
            if torch.sum(self.pbc) > 0.0:
                raise SystemException("Found periodic boundary conditions but no cell.")
            else:
                self.cells = None

        # 6) Do proper broadcasting here for easier use in e.g. integrators and
        #    thermostats afterwards
        self.masses = self.masses[None, :, :, None]
        self.atom_masks = self.atom_masks[..., None]

        # 7) Initialize Momenta
        if self.initializer:
            self.initializer.initialize_system(self)

    @property
    def center_of_mass(self):
        """
        Compute the center of mass for each replica and molecule

        Returns:
            torch.Tensor: n_replicas x n_molecules x 1 x 3 tensor holding the
                          center of mass.
        """
        # Mask mass array
        masses = self.masses * self.atom_masks
        # Compute center of mass
        center_of_mass = torch.sum(self.positions * masses, 2) / torch.sum(masses, 2)
        return center_of_mass

    def remove_com(self):
        """
        Move all structures to their respective center of mass.
        """
        # Mask to avoid offsets
        self.positions -= self.center_of_mass[:, :, None, :]
        # Apply atom masks to avoid artificial shifts
        self.positions *= self.atom_masks

    def remove_com_translation(self):
        """
        Remove all components in the current momenta associated with
        translational motion.
        """
        self.momenta -= (
            torch.sum(self.momenta, 2, keepdim=True)
            / self.n_atoms.float()[None, :, None, None]
        )
        # Apply atom masks to avoid artificial shifts
        self.momenta *= self.atom_masks

    def remove_com_rotation(self, detach=True):
        """
        Remove all components in the current momenta associated with rotational
        motion using Eckart conditons.

        Args:
            detach (bool): Whether computational graph should be detached in
                           order to accelerated the simulation (default=True).
        """
        # Compute the moment of inertia tensor
        moment_of_inertia = (
            torch.sum(self.positions ** 2, 3, keepdim=True)[..., None]
            * torch.eye(3, device=self.device)[None, None, None, :, :]
            - self.positions[..., :, None] * self.positions[..., None, :]
        )
        moment_of_inertia = torch.sum(moment_of_inertia * self.masses[..., None], 2)

        # Compute the angular momentum
        angular_momentum = torch.sum(torch.cross(self.positions, self.momenta, -1), 2)

        # Compute the angular velocities
        angular_velocities = torch.matmul(
            angular_momentum[:, :, None, :], batch_inverse(moment_of_inertia)
        )

        # Compute individual atomic contributions
        rotational_velocities = torch.cross(
            angular_velocities.repeat(1, 1, self.max_n_atoms, 1), self.positions, -1
        )

        if detach:
            rotational_velocities = rotational_velocities.detach()

        # Subtract rotation from overall motion (apply atom mask)
        self.momenta -= rotational_velocities * self.masses * self.atom_masks

    def wrap_positions(self, eps=1e-6):
        """
        Move atoms outside the box back into the box for all dimensions with periodic boundary
        conditions.

        Args:
            eps (float): Small offset for numerical stability
        """
        # Compute fractional coordinates
        with torch.no_grad():
            tmp_positions = self.positions.transpose(2, 3)
            tmp_cells = self.cells.transpose(2, 3)
            inv_positions, _ = torch.solve(tmp_positions, tmp_cells)
            inv_positions = inv_positions.transpose(2, 3)

            # Get periodic coordinates
            periodic = torch.masked_select(inv_positions, self.pbc[None, :, None, :])

            # Apply periodic boundary conditions (with small buffer)
            periodic = periodic + eps
            periodic = periodic % 1.0
            periodic = periodic - eps

            # Update fractional coordinates
            inv_positions.masked_scatter_(self.pbc[None, :, None, :], periodic)

            # Convert to positions
            self.positions = torch.matmul(inv_positions, self.cells)

    def compute_pressure(self, tensor=False, kinetic_component=False):
        """
        Compute the pressure (tensor) based on the stress tensor of the systems.

        Args:
            tensor (bool): Instead of a scalar pressure, return the full pressure tensor. (Required for
                           anisotropic cell deformation.)
            kinetic_component (bool): Include the kinetic energy component during the computation of the
                                      pressure (default=False).

        Returns:
            torch.Tensor: Depending on the tensor-flag, returns a tensor containing the pressure with dimensions
                          n_replicas x n_molecules (False) or n_replicas x n_molecules x 3 x 3 (True).
        """
        if self.stress is None:
            raise SystemError(
                "Stress required for computation of the instantaneous pressure."
            )

        if tensor:
            pressure = -self.stress
            if kinetic_component:
                pressure = (
                    pressure + 2.0 * self.kinetic_energy_tensor / self.volume[..., None]
                )

        else:
            # Einsum computes the trace
            pressure = -torch.einsum("abii->ab", self.stress) / 3.0

            if kinetic_component:
                pressure = pressure + 2.0 * self.kinetic_energy / self.volume / 3.0

        return pressure.detach()

    def get_ase_atoms(self, internal_units=True):
        """
        Convert the stored molecular configurations into ASE Atoms objects. This is e.g. used for the
        neighbor lists based on environment providers. All units are atomic units by default, as used in the calculator

        Args:
            internal_units (bool): Whether atomic units or Angstrom should be returned (default=True). This should
                                 always be True when used with an EnvironmentProviderNeighborList.

        Returns:
            list(ase.Atoms): List of ASE Atoms objects, with the replica and molecule dimension flattened.
        """
        atoms = []
        for idx_r in range(self.n_replicas):
            for idx_m in range(self.n_molecules):
                positions = (
                    self.positions[idx_r, idx_m, : self.n_atoms[idx_m]]
                    .cpu()
                    .detach()
                    .numpy()
                )
                atom_types = (
                    self.atom_types[idx_r, idx_m, : self.n_atoms[idx_m]]
                    .cpu()
                    .detach()
                    .numpy()
                )
                if self.cells is not None:
                    cell = self.cells[idx_r, idx_m].cpu().detach().numpy()
                    pbc = self.pbc[idx_m].cpu().detach().numpy()
                else:
                    cell = None
                    pbc = None

                # If requested, convert units of space to Angstrom
                if not internal_units:
                    positions /= MDUnits.angs2internal
                    if cell is not None:
                        cell /= MDUnits.angs2internal

                mol = Atoms(atom_types, positions, cell=cell, pbc=pbc)
                atoms.append(mol)

        return atoms

    @property
    def velocities(self):
        """
        Convenience property to access molecular velocities instead of the
        momenta (e.g for power spectra)

        Returns:
            torch.Tensor: Velocity tensor with the same shape as the momenta.
        """
        return self.momenta / self.masses

    @property
    def centroid_positions(self):
        """
        Convenience property to access the positions of the centroid during
        ring polymer molecular dynamics. Does not make sense during a
        standard dynamics setup.

        Returns:
            torch.Tensor: Tensor of the shape 1 x n_molecules x n_atoms x 3
            holding the centroid positions.
        """
        return compute_centroid(self.positions)

    @property
    def centroid_momenta(self):
        """
        Convenience property to access the centroid momenta during ring
        polymer molecular dynamics. Does not make sense during a standard
        dynamics setup.

        Returns:
            torch.Tensor: Tensor of the shape 1 x n_molecules x n_atoms x 3
                          holding the centroid momenta.
        """
        return compute_centroid(self.momenta)

    @property
    def centroid_velocities(self):
        """
        Convenience property to access the velocities of the centroid during
        ring polymer molecular dynamics (e.g. for computing power spectra).
        Does not make sense during a standard dynamics setup.

        Returns:
            torch.Tensor: Tensor of the shape 1 x n_molecules x n_atoms x 3
            holding the centroid velocities.
        """
        return self.centroid_momenta / self.masses

    @property
    def kinetic_energy(self):
        """
        Convenience property for computing the kinetic energy associated with
        each replica and molecule.

        Returns:
            torch.Tensor: Tensor of the kinetic energies (in Hartree) with
                          the shape n_replicas x n_molecules
        """
        # Apply atom mask
        momenta = self.momenta * self.atom_masks

        kinetic_energy = 0.5 * torch.sum(
            torch.sum(momenta ** 2, 3) / self.masses[..., 0], 2
        )
        return kinetic_energy.detach()

    @property
    def kinetic_energy_tensor(self):
        """
        Compute the kinetic energy tensor (outer product of momenta divided by masses) for pressure computation.
        The standard kinetic energy is the trace of this tensor.

        Returns:
            torch.tensor: n_replicas x n_molecules x 3 x 3 tensor containg kinetic energy components.

        """
        # Apply atom mask
        momenta = self.momenta * self.atom_masks
        kinetic_energy_tensor = 0.5 * torch.sum(
            momenta[..., None] * momenta[:, :, :, None, :] / self.masses[..., None], 2
        )
        return kinetic_energy_tensor.detach()

    @property
    def temperature(self):
        """
        Convenience property for accessing the instantaneous temperatures of
        each replica and molecule.

        Returns:
            torch.Tensor: Tensor of the instantaneous temperatures (in
                          Kelvin) with the shape n_replicas x n_molecules
        """
        temperature = (
            2.0
            / (3.0 * MDUnits.kB * self.n_atoms.float()[None, :])
            * self.kinetic_energy
        )
        return temperature

    @property
    def centroid_kinetic_energy(self):
        """
        Convenience property for computing the kinetic energy associated with
        the centroid of each molecule. Only sensible in the context of ring
        polymer molecular dynamics.

        Returns:
            torch.Tensor: Tensor of the centroid kinetic energies (in
                          Hartree) with the shape 1 x n_molecules
        """
        # Apply atom mask
        centroid_momenta = self.centroid_momenta * self.atom_masks
        kinetic_energy = 0.5 * torch.sum(
            torch.sum(centroid_momenta ** 2, 3) / self.masses[..., 0], 2
        )
        return kinetic_energy

    @property
    def centroid_temperature(self):
        """
        Convenience property for accessing the instantaneous temperatures of
        the centroid of each molecule. Only makes sense in the context of
        ring polymer molecular dynamics.

        Returns:
            torch.Tensor: Tensor of the instantaneous centroid temperatures (
                          in Kelvin) with the shape 1 x n_molecules
        """
        temperature = (
            2.0
            / (3.0 * self.n_atoms.float()[None, :] * MDUnits.kB)
            * self.centroid_kinetic_energy
        )
        return temperature

    @property
    def volume(self):
        """
        Compute the cell volumes if cells are present.

        Returns:
            torch.tensor: n_replicas x n_molecules containing the volumes.
        """
        if self.cells is None:
            return None
        else:
            volume = torch.sum(
                self.cells[:, :, 0]
                * torch.cross(self.cells[:, :, 1], self.cells[:, :, 2], dim=2),
                dim=2,
                keepdim=False,
            )
            return volume.detach()

    @property
    def state_dict(self):
        """
        State dict for storing the system state.

        Returns:
            dict: Dictionary containing all properties for restoring the
                  current state of the system during simulation.
        """
        state_dict = {
            "positions": self.positions,
            "momenta": self.momenta,
            "forces": self.forces,
            "properties": self.properties,
            "n_atoms": self.n_atoms,
            "atom_types": self.atom_types,
            "masses": self.masses,
            "cells": self.cells,
            "pbc": self.pbc,
            "stress": self.stress,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        """
        Routine for restoring the state of a system specified in a previously
        stored state dict. Used to restart molecular dynamics simulations.

        Args:
            state_dict (dict): State dict of the system state.
        """
        self.positions = state_dict["positions"]
        self.momenta = state_dict["momenta"]
        self.forces = state_dict["forces"]
        self.properties = state_dict["properties"]
        self.n_atoms = state_dict["n_atoms"]
        self.atom_types = state_dict["atom_types"]
        self.masses = state_dict["masses"]
        self.cells = state_dict["cells"]
        self.pbc = state_dict["pbc"]
        self.stress = state_dict["stress"]

        self.n_replicas = self.positions.shape[0]
        self.n_molecules = self.positions.shape[1]
        self.max_n_atoms = self.positions.shape[2]

        # Build atom masks according to the present number of atoms
        for i in range(self.n_molecules):
            self.atom_masks[:, i, : self.n_atoms[i], :] = 1.0



r"""
Classes to manage the training process.

schnetpack.train.Trainer encapsulates the training loop. It also can automatically monitor the performance on the
validation set and contains logic for checkpointing. The training process can be customized using Hooks which derive
from schnetpack.train.Hooks.

"""

from .trainer import Trainer
from .loss import *
from .hooks import *
from .metrics import *



import numpy as np
import torch
import os
import schnetpack as spk
from torch.utils.data.sampler import RandomSampler
from schnetpack.utils.script_utils.script_error import ScriptError


__all__ = ["get_loaders", "get_statistics", "get_dataset"]


def get_statistics(
    args, split_path, train_loader, atomref, divide_by_atoms=False, logging=None
):
    """
    Get statistics for molecular properties. Use split file if possible.

    Args:
        args (argparse.Namespace): parsed script arguments
        split_path (str): path to the split file
        train_loader (spk.data.AtomsLoader): dataloader for training set
        atomref (dict): atomic references
        divide_by_atoms (dict or bool): divide mean by number of atoms if True
        logging: logger

    Returns:
        mean (dict): mean values for the selected properties
        stddev (dict): stddev values for the selected properties
    """
    # check if split file exists
    if not os.path.exists(split_path):
        raise ScriptError("No split file found ad {}".format(split_path))
    split_data = np.load(split_path)

    # check if split file contains statistical data
    if "mean" in split_data.keys():
        mean = {args.property: torch.from_numpy(split_data["mean"])}
        stddev = {args.property: torch.from_numpy(split_data["stddev"])}
        if logging is not None:
            logging.info("cached statistics was loaded...")

    # calculate statistical data
    else:
        mean, stddev = train_loader.get_statistics(
            args.property, divide_by_atoms, atomref
        )
        np.savez(
            split_path,
            train_idx=split_data["train_idx"],
            val_idx=split_data["val_idx"],
            test_idx=split_data["test_idx"],
            mean=mean[args.property].numpy(),
            stddev=stddev[args.property].numpy(),
        )

    return mean, stddev


def get_loaders(args, dataset, split_path, logging=None):
    """
    Create train-val-test splits for dataset and return the corresponding dataloaders.

    Args:
        args (argparse.Namespace): parsed script arguments
        dataset (spk.AtomsData): total dataset
        split_path (str): path to split file
        logging: logger

    Returns:
        (spk.AtomsLoader, spk.AtomsLoader, spk.AtomsLoader): dataloaders for train,
            val and test
    """
    if logging is not None:
        logging.info("create splits...")

    # create or load dataset splits depending on args.mode
    if args.mode == "train":
        data_train, data_val, data_test = spk.data.train_test_split(
            dataset, *args.split, split_file=split_path
        )
    else:
        data_train, data_val, data_test = spk.data.train_test_split(
            dataset, split_file=split_path
        )

    if logging is not None:
        logging.info("load data...")

    # build dataloaders
    train_loader = spk.data.AtomsLoader(
        data_train,
        batch_size=args.batch_size,
        sampler=RandomSampler(data_train),
        num_workers=4,
        pin_memory=args.cuda,
    )
    val_loader = spk.data.AtomsLoader(
        data_val, batch_size=args.batch_size, num_workers=2, pin_memory=args.cuda
    )
    test_loader = spk.data.AtomsLoader(
        data_test, batch_size=args.batch_size, num_workers=2, pin_memory=args.cuda
    )

    return train_loader, val_loader, test_loader


def get_dataset(args, environment_provider, logging=None):
    """
    Get dataset from arguments.

    Args:
        args (argparse.Namespace): parsed arguments
        environment_provider (spk.environment.BaseEnvironmentProvider): environment-
            provider of dataset
        logging: logger

    Returns:
        spk.data.AtomsData: dataset

    """
    if args.dataset == "qm9":
        if logging:
            logging.info("QM9 will be loaded...")
        qm9 = spk.datasets.QM9(
            args.datapath,
            download=True,
            load_only=[args.property],
            collect_triples=args.model == "wacsf",
            remove_uncharacterized=args.remove_uncharacterized,
            environment_provider=environment_provider,
        )
        return qm9
    elif args.dataset == "ani1":
        if logging:
            logging.info("ANI1 will be loaded...")
        ani1 = spk.datasets.ANI1(
            args.datapath,
            download=True,
            load_only=[args.property],
            collect_triples=args.model == "wacsf",
            num_heavy_atoms=args.num_heavy_atoms,
            environment_provider=environment_provider,
        )
        return ani1
    elif args.dataset == "md17":
        if logging:
            logging.info("MD17 will be loaded...")
        md17 = spk.datasets.MD17(
            args.datapath,
            args.molecule,
            download=True,
            collect_triples=args.model == "wacsf",
            environment_provider=environment_provider,
        )
        return md17
    elif args.dataset == "matproj":
        if logging:
            logging.info("Materials project will be loaded...")
        mp = spk.datasets.MaterialsProject(
            args.datapath,
            apikey=args.apikey,
            download=True,
            load_only=[args.property],
            environment_provider=environment_provider,
        )
        if args.timestamp:
            mp = mp.at_timestamp(args.timestamp)
        return mp
    elif args.dataset == "omdb":
        if logging:
            logging.info("Organic Materials Database will be loaded...")
        omdb = spk.datasets.OrganicMaterialsDatabase(
            args.datapath,
            download=True,
            load_only=[args.property],
            environment_provider=environment_provider,
        )
        return omdb
    elif args.dataset == "custom":
        if logging:
            logging.info("Custom dataset will be loaded...")

        # define properties to be loaded
        load_only = [args.property]
        if args.derivative is not None:
            load_only.append(args.derivative)

        if args.stress is not None:
            load_only.append(args.stress)

        dataset = spk.AtomsData(
            args.datapath,
            load_only=load_only,
            collect_triples=args.model == "wacsf",
            environment_provider=environment_provider,
        )
        return dataset
    else:
        raise spk.utils.ScriptError("Invalid dataset selected!")



import torch
from schnetpack import representation

__all__ = ["compute_centroid", "batch_inverse", "check_triples_required"]


def compute_centroid(ensemble):
    """
    Compute centroids of the system properties (e.g. positions, momenta) given in ensemble with respect to the replica
    dimension (0). The original dimensionality of the tensor is kept for the purpose of broadcasting and logging. This
    routine is primarily intended to be used for ring polymer simulations.

    Args:
        ensemble (torch.Tensor): System property tensor (e.g. positions, momenta) of the general dimension
                                 n_replicas x n_molecules x ...

    Returns:
        torch.Tensor: Centroid averaged over the replica dimension with the general shape 1 x n_molecules x ...
    """
    centroid = torch.mean(ensemble, 0, keepdim=True)
    return centroid


def batch_inverse(tensor):
    """
    Compute the matrix inverse of a batch of square matrices. This routine is used for removing rotational motion
    during the molecular dynamics simulation.

    Args:
        tensor (torch.Tensor):  Tensor of square matrices with the shape n_batch x dim1 x dim1

    Returns:
        torch.Tensor: Tensor of the inverted square matrices with the same shape as the input tensor.
    """
    return torch.linalg.inv(tensor)


class RunningAverage:
    """
    Running average class for logging purposes. Accumulates the average of a given tensor over the course of the
    simulation.
    """

    def __init__(self):
        # Initialize running average and item count
        self.average = 0
        self.counts = 0

    def update(self, value):
        """
        Update the running average.

        Args:
            value (torch.Tensor): Tensor containing the property whose average should be accumulated.
        """
        self.average = (self.counts * self.average + value) / (self.counts + 1)
        self.counts += 1


def check_triples_required(model):
    """
    Check, if the model has a representation which requires the computation of triples between atoms.
    E.g. angular terms in Behler symmetry functions.
    Args:
        model (schnetpack.model.AtomisticModel):
    Returns:
        bool: Indicator whether model needs triples or not.
    """
    # check if DataParallel or conventional
    if hasattr(model, "module"):
        representation_layer = model.module.representation
    else:
        representation_layer = model.representation

    # If representation is standardized, extract it from standardization layer
    if isinstance(representation_layer, representation.StandardizeSF):
        representation_layer = representation_layer.SFBlock

    # Check if representation uses SFs
    if isinstance(representation_layer, representation.SymmetryFunctions):
        # Check if angular terms are present:
        if representation_layer.n_angular > 0:
            return True
        else:
            return False
    else:
        return False



import os
import sys
import numpy as np
import torch


class Trainer:
    r"""Class to train a model.

    This contains an internal training loop which takes care of validation and can be
    extended with custom functionality using hooks.

    Args:
       model_path (str): path to the model directory.
       model (torch.Module): model to be trained.
       loss_fn (callable): training loss function.
       optimizer (torch.optim.optimizer.Optimizer): training optimizer.
       train_loader (torch.utils.data.DataLoader): data loader for training set.
       validation_loader (torch.utils.data.DataLoader): data loader for validation set.
       keep_n_checkpoints (int, optional): number of saved checkpoints.
       checkpoint_interval (int, optional): intervals after which checkpoints is saved.
       hooks (list, optional): hooks to customize training process.
       loss_is_normalized (bool, optional): if True, the loss per data point will be
           reported. Otherwise, the accumulated loss is reported.

    """

    def __init__(
        self,
        model_path,
        model,
        loss_fn,
        optimizer,
        train_loader,
        validation_loader,
        keep_n_checkpoints=3,
        checkpoint_interval=10,
        validation_interval=1,
        hooks=[],
        loss_is_normalized=True,
    ):
        self.model_path = model_path
        self.checkpoint_path = os.path.join(self.model_path, "checkpoints")
        self.best_model = os.path.join(self.model_path, "best_model")
        self.train_loader = train_loader
        self.validation_loader = validation_loader
        self.validation_interval = validation_interval
        self.keep_n_checkpoints = keep_n_checkpoints
        self.hooks = hooks
        self.loss_is_normalized = loss_is_normalized

        self._model = model
        self._stop = False
        self.checkpoint_interval = checkpoint_interval

        self.loss_fn = loss_fn
        self.optimizer = optimizer

        if os.path.exists(self.checkpoint_path):
            self.restore_checkpoint()
        else:
            os.makedirs(self.checkpoint_path)
            self.epoch = 0
            self.step = 0
            self.best_loss = float("inf")
            self.store_checkpoint()

    def _check_is_parallel(self):
        return True if isinstance(self._model, torch.nn.DataParallel) else False

    def _load_model_state_dict(self, state_dict):
        if self._check_is_parallel():
            self._model.module.load_state_dict(state_dict)
        else:
            self._model.load_state_dict(state_dict)

    def _optimizer_to(self, device):
        """
        Move the optimizer tensors to device before training.

        Solves restore issue:
        https://github.com/atomistic-machine-learning/schnetpack/issues/126
        https://github.com/pytorch/pytorch/issues/2830

        """
        for state in self.optimizer.state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(device)

    @property
    def state_dict(self):
        state_dict = {
            "epoch": self.epoch,
            "step": self.step,
            "best_loss": self.best_loss,
            "optimizer": self.optimizer.state_dict(),
            "hooks": [h.state_dict for h in self.hooks],
        }
        if self._check_is_parallel():
            state_dict["model"] = self._model.module.state_dict()
        else:
            state_dict["model"] = self._model.state_dict()
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.epoch = state_dict["epoch"]
        self.step = state_dict["step"]
        self.best_loss = state_dict["best_loss"]
        self.optimizer.load_state_dict(state_dict["optimizer"])
        self._load_model_state_dict(state_dict["model"])

        for h, s in zip(self.hooks, state_dict["hooks"]):
            h.state_dict = s

    def store_checkpoint(self):
        chkpt = os.path.join(
            self.checkpoint_path, "checkpoint-" + str(self.epoch) + ".pth.tar"
        )
        torch.save(self.state_dict, chkpt)

        chpts = [f for f in os.listdir(self.checkpoint_path) if f.endswith(".pth.tar")]
        if len(chpts) > self.keep_n_checkpoints:
            chpt_epochs = [int(f.split(".")[0].split("-")[-1]) for f in chpts]
            sidx = np.argsort(chpt_epochs)
            for i in sidx[: -self.keep_n_checkpoints]:
                os.remove(os.path.join(self.checkpoint_path, chpts[i]))

    def restore_checkpoint(self, epoch=None):
        if epoch is None:
            epoch = max(
                [
                    int(f.split(".")[0].split("-")[-1])
                    for f in os.listdir(self.checkpoint_path)
                    if f.startswith("checkpoint")
                ]
            )

        chkpt = os.path.join(
            self.checkpoint_path, "checkpoint-" + str(epoch) + ".pth.tar"
        )
        self.state_dict = torch.load(chkpt)

    def train(self, device, n_epochs=sys.maxsize):
        """Train the model for the given number of epochs on a specified device.

        Args:
            device (torch.torch.Device): device on which training takes place.
            n_epochs (int): number of training epochs.

        Note: Depending on the `hooks`, training can stop earlier than `n_epochs`.

        """
        self._model.to(device)
        self._optimizer_to(device)
        self._stop = False

        for h in self.hooks:
            h.on_train_begin(self)

        try:
            for _ in range(n_epochs):
                # increase number of epochs by 1
                self.epoch += 1

                for h in self.hooks:
                    h.on_epoch_begin(self)

                if self._stop:
                    # decrease self.epoch if training is aborted on epoch begin
                    self.epoch -= 1
                    break

                # perform training epoch
                #                if progress:
                #                    train_iter = tqdm(self.train_loader)
                #                else:
                train_iter = self.train_loader

                self._model.train()
                for train_batch in train_iter:
                    self.optimizer.zero_grad()

                    for h in self.hooks:
                        h.on_batch_begin(self, train_batch)

                    # move input to gpu, if needed
                    train_batch = {k: v.to(device) for k, v in train_batch.items()}

                    result = self._model(train_batch)
                    loss = self.loss_fn(train_batch, result)
                    loss.backward()
                    self.optimizer.step()
                    self.step += 1

                    for h in self.hooks:
                        h.on_batch_end(self, train_batch, result, loss)

                    if self._stop:
                        break

                if self.epoch % self.checkpoint_interval == 0:
                    self.store_checkpoint()

                # validation
                self._model.eval()
                if self.epoch % self.validation_interval == 0 or self._stop:
                    for h in self.hooks:
                        h.on_validation_begin(self)

                    val_loss = 0.0
                    n_val = 0
                    for val_batch in self.validation_loader:
                        # append batch_size
                        vsize = list(val_batch.values())[0].size(0)
                        n_val += vsize

                        for h in self.hooks:
                            h.on_validation_batch_begin(self)

                        # move input to gpu, if needed
                        val_batch = {k: v.to(device) for k, v in val_batch.items()}

                        val_result = self._model(val_batch)
                        val_batch_loss = (
                            self.loss_fn(val_batch, val_result).data.cpu().numpy()
                        )
                        if self.loss_is_normalized:
                            val_loss += val_batch_loss * vsize
                        else:
                            val_loss += val_batch_loss

                        for h in self.hooks:
                            h.on_validation_batch_end(self, val_batch, val_result)

                    # weighted average over batches
                    if self.loss_is_normalized:
                        val_loss /= n_val

                    if self.best_loss > val_loss:
                        self.best_loss = val_loss
                        torch.save(self._model, self.best_model)

                    for h in self.hooks:
                        h.on_validation_end(self, val_loss)

                for h in self.hooks:
                    h.on_epoch_end(self)

                if self._stop:
                    break
            #
            # Training Ends
            #
            # run hooks & store checkpoint
            for h in self.hooks:
                h.on_train_ends(self)
            self.store_checkpoint()

        except Exception as e:
            for h in self.hooks:
                h.on_train_failed(self)

            raise e



import numpy as np
import torch
from torch import nn


__all__ = ["CosineCutoff", "MollifierCutoff", "HardCutoff", "get_cutoff_by_string"]


def get_cutoff_by_string(key):
    # build cutoff module
    if key == "hard":
        cutoff_network = HardCutoff
    elif key == "cosine":
        cutoff_network = CosineCutoff
    elif key == "mollifier":
        cutoff_network = MollifierCutoff
    else:
        raise NotImplementedError("cutoff_function {} is unknown".format(key))
    return cutoff_network


class CosineCutoff(nn.Module):
    r"""Class of Behler cosine cutoff.

    .. math::
       f(r) = \begin{cases}
        0.5 \times \left[1 + \cos\left(\frac{\pi r}{r_\text{cutoff}}\right)\right]
          & r < r_\text{cutoff} \\
        0 & r \geqslant r_\text{cutoff} \\
        \end{cases}

    Args:
        cutoff (float, optional): cutoff radius.

    """

    def __init__(self, cutoff=5.0):
        super(CosineCutoff, self).__init__()
        self.register_buffer("cutoff", torch.FloatTensor([cutoff]))

    def forward(self, distances):
        """Compute cutoff.

        Args:
            distances (torch.Tensor): values of interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        # Compute values of cutoff function
        cutoffs = 0.5 * (torch.cos(distances * np.pi / self.cutoff) + 1.0)
        # Remove contributions beyond the cutoff radius
        cutoffs *= (distances < self.cutoff).float()
        return cutoffs


class MollifierCutoff(nn.Module):
    r"""Class for mollifier cutoff scaled to have a value of 1 at :math:`r=0`.

    .. math::
       f(r) = \begin{cases}
        \exp\left(1 - \frac{1}{1 - \left(\frac{r}{r_\text{cutoff}}\right)^2}\right)
          & r < r_\text{cutoff} \\
        0 & r \geqslant r_\text{cutoff} \\
        \end{cases}

    Args:
        cutoff (float, optional): Cutoff radius.
        eps (float, optional): offset added to distances for numerical stability.

    """

    def __init__(self, cutoff=5.0, eps=1.0e-7):
        super(MollifierCutoff, self).__init__()
        self.register_buffer("cutoff", torch.FloatTensor([cutoff]))
        self.register_buffer("eps", torch.FloatTensor([eps]))

    def forward(self, distances):
        """Compute cutoff.

        Args:
            distances (torch.Tensor): values of interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        mask = (distances + self.eps < self.cutoff).float()
        exponent = 1.0 - 1.0 / (1.0 - torch.pow(distances * mask / self.cutoff, 2))
        cutoffs = torch.exp(exponent)
        cutoffs = cutoffs * mask
        return cutoffs


class HardCutoff(nn.Module):
    r"""Class of hard cutoff.

    .. math::
       f(r) = \begin{cases}
        1 & r \leqslant r_\text{cutoff} \\
        0 & r > r_\text{cutoff} \\
        \end{cases}

    Args:
        cutoff (float): cutoff radius.

    """

    def __init__(self, cutoff=5.0):
        super(HardCutoff, self).__init__()
        self.register_buffer("cutoff", torch.FloatTensor([cutoff]))

    def forward(self, distances):
        """Compute cutoff.

        Args:
            distances (torch.Tensor): values of interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        mask = (distances <= self.cutoff).float()
        return mask



import numpy as np
import torch

__all__ = ["NormalModeTransformer"]


class NormalModeTransformer:
    """
    Class for transforming between bead and normal mode representation of the ring polymer, used e.g. in propagating the
    ring polymer during simulation. An in depth description of the transformation can be found e.g. in [#rpmd3]_. Here,
    a simple matrix multiplication is used instead of a Fourier transformation, which can be more performant in certain
    cases. On the GPU however, no significant performance gains where observed when using a FT based transformation over
    the matrix version.

    This transformation operates on the first dimension of the property tensors (e.g. positions, momenta) defined in the
    system class. Hence, the transformation can be carried out for several molecules at the same time.

    Args:
        n_beads (int): Number of beads in the ring polymer.
        device (str): Computation device (default='cuda').

    References
    ----------
    .. [#rpmd3] Ceriotti, Parrinello, Markland, Manolopoulos:
       Efficient stochastic thermostatting of path integral molecular dynamics.
       The Journal of Chemical Physics, 133, 124105. 2010.
    """

    def __init__(self, n_beads, device="cuda"):
        self.n_beads = n_beads

        self.device = device

        # Initialize the transformation matrix
        self.c_transform = self._init_transformation_matrix()

    def _init_transformation_matrix(self):
        """
        Build the normal mode transformation matrix. This matrix only has to be built once and can then be used during
        the whole simulation. The matrix has the dimension n_beads x n_beads, where n_beads is the number of beads in
        the ring polymer

        Returns:
            torch.Tensor: Normal mode transformation matrix of the shape n_beads x n_beads
        """
        # Set up basic transformation matrix
        c_transform = np.zeros((self.n_beads, self.n_beads))

        # Get auxiliary array with bead indices
        n = np.arange(1, self.n_beads + 1)

        # for k = 0
        c_transform[0, :] = 1.0

        for k in range(1, self.n_beads // 2 + 1):
            c_transform[k, :] = np.sqrt(2) * np.cos(2 * np.pi * k * n / self.n_beads)

        for k in range(self.n_beads // 2 + 1, self.n_beads):
            c_transform[k, :] = np.sqrt(2) * np.sin(2 * np.pi * k * n / self.n_beads)

        if self.n_beads % 2 == 0:
            c_transform[self.n_beads // 2, :] = (-1) ** n

        # Since matrix is initialized as C(k,n) does not need to be transposed
        c_transform /= np.sqrt(self.n_beads)
        c_transform = torch.from_numpy(c_transform).float().to(self.device)

        return c_transform

    def beads2normal(self, x_beads):
        """
        Transform a system tensor (e.g. momenta, positions) from the bead representation to normal mode representation.

        Args:
            x_beads (torch.Tensor): System tensor in bead representation with the general shape
                                    n_beads x n_molecules x ...

        Returns:
            torch.Tensor: System tensor in normal mode representation with the same shape as the input tensor.
        """
        return torch.mm(self.c_transform, x_beads.view(self.n_beads, -1)).view(
            x_beads.shape
        )

    def normal2beads(self, x_normal):
        """
        Transform a system tensor (e.g. momenta, positions) in normal mode representation back to bead representation.

        Args:
            x_normal (torch.Tensor): System tensor in normal mode representation with the general shape
                                    n_beads x n_molecules x ...

        Returns:
            torch.Tensor: System tensor in bead representation with the same shape as the input tensor.
        """
        return torch.mm(
            self.c_transform.transpose(0, 1), x_normal.view(self.n_beads, -1)
        ).view(x_normal.shape)



import torch


__all__ = ["build_mse_loss"]


class LossFnError(Exception):
    pass


def build_mse_loss(properties, loss_tradeoff=None):
    """
    Build the mean squared error loss function.

    Args:
        properties (list): mapping between the model properties and the
            dataset properties
        loss_tradeoff (list or None): multiply loss value of property with tradeoff
            factor

    Returns:
        mean squared error loss function

    """
    if loss_tradeoff is None:
        loss_tradeoff = [1] * len(properties)
    if len(properties) != len(loss_tradeoff):
        raise LossFnError("loss_tradeoff must have same length as properties!")

    def loss_fn(batch, result):
        loss = 0.0
        for prop, factor in zip(properties, loss_tradeoff):
            diff = batch[prop] - result[prop]
            diff = diff ** 2
            err_sq = factor * torch.mean(diff)
            loss += err_sq
        return loss

    return loss_fn



import numpy as np
from ase.db import connect
from ase import Atoms
from schnetpack import Properties


class Evaluator:
    def __init__(self, model, dataloader):
        """
        Base class for model predictions.

        Args:
            model (torch.nn.Module): trained model
        """
        self.model = model
        self.dataloader = dataloader

    def _get_predicted(self, device):
        """
        Calculate the predictions for the dataloader.

        Args:
            device (str): cpu or cuda

        Returns:

        """
        self.model.eval()

        predicted = {}
        for batch in self.dataloader:
            # build batch for prediction
            batch = {k: v.to(device) for k, v in batch.items()}
            # predict
            result = self.model(batch)
            # store prediction batches to dict
            for p in result.keys():
                value = result[p].cpu().detach().numpy()
                if p in predicted.keys():
                    predicted[p].append(value)
                else:
                    predicted[p] = [value]
            # store positions, numbers and mask to dict
            for p in [Properties.R, Properties.Z, Properties.atom_mask]:
                value = batch[p].cpu().detach().numpy()
                if p in predicted.keys():
                    predicted[p].append(value)
                else:
                    predicted[p] = [value]

        max_shapes = {
            prop: max([list(val.shape) for val in values])
            for prop, values in predicted.items()
        }
        for prop, values in predicted.items():
            max_shape = max_shapes[prop]
            predicted[prop] = np.vstack(
                [
                    np.lib.pad(
                        batch,
                        [
                            [0, add_dims]
                            for add_dims in max_shape - np.array(batch.shape)
                        ],
                        mode="constant",
                    )
                    for batch in values
                ]
            )

        return predicted

    def evaluate(self, device):
        raise NotImplementedError


class NPZEvaluator(Evaluator):
    def __init__(self, model, dataloader, out_file):
        self.out_file = out_file
        super(NPZEvaluator, self).__init__(model=model, dataloader=dataloader)

    def evaluate(self, device):
        predicted = self._get_predicted(device)
        np.savez(self.out_file, **predicted)


class DBEvaluator(Evaluator):
    def __init__(self, model, dataloader, out_file):
        self.dbpath = dataloader.dataset.dbpath
        self.out_file = out_file
        super(DBEvaluator, self).__init__(model=model, dataloader=dataloader)

    def evaluate(self, device):
        predicted = self._get_predicted(device)
        positions = predicted.pop(Properties.R)
        atomic_numbers = predicted.pop(Properties.Z)
        atom_masks = predicted.pop(Properties.atom_mask).astype(bool)
        with connect(self.out_file) as conn:
            for i, mask in enumerate(atom_masks):
                z = atomic_numbers[i, mask]
                r = positions[i, mask]
                ats = Atoms(numbers=z, positions=r)
                data = {
                    prop: self._unpad(mask, values[i])
                    for prop, values in predicted.items()
                }
                conn.write(ats, data=data)

    def _unpad(self, mask, values):
        if len(values.shape) == 1:
            return values
        return values[mask]



import os
import tarfile

import numpy as np
from ase.io import read
from ase.db import connect
from ase.units import eV

import schnetpack as spk
from schnetpack.datasets import DownloadableAtomsData


__all__ = ["OrganicMaterialsDatabase"]


class OrganicMaterialsDatabase(DownloadableAtomsData):
    """Organic Materials Database (OMDB) of bulk organic crystals.

    Registration to the OMDB is free for academic users. This database contains DFT
    (PBE) band gap (OMDB-GAP1 database) for 12500 non-magnetic materials.

    Args:
        path (str): path to directory containing database.
        cutoff (float): cutoff for bulk interactions.
        download (bool, optional): enable downloading if database does not exists.
        subset (list, optional): Deprecated! Do not use! Subsets are created with
            AtomsDataSubset class.
        load_only (list, optional): reduced set of properties to be loaded
        collect_triples (bool, optional): Set to True if angular features are needed.
        environment_provider (spk.environment.BaseEnvironmentProvider): define how
            neighborhood is calculated
            (default=spk.environment.SimpleEnvironmentProvider).

    References:
        Olsthoorn, B., Geilhufe, R.M., Borysov, S.S. and Balatsky, A.V. (2019),
        Band Gap Prediction for Large Organic Crystal Structures with Machine Learning.
        Adv. Quantum Technol., 2: 1900023. https://doi.org/10.1002/qute.201900023

    """

    BandGap = "band_gap"

    def __init__(
        self,
        path,
        download=True,
        subset=None,
        load_only=None,
        collect_triples=False,
        environment_provider=spk.environment.SimpleEnvironmentProvider(),
    ):
        available_properties = [OrganicMaterialsDatabase.BandGap]

        units = [eV]

        self.path = path

        dbpath = self.path.replace(".tar.gz", ".db")
        self.dbpath = dbpath

        if not os.path.exists(path) and not os.path.exists(dbpath):
            raise FileNotFoundError(
                "Download OMDB dataset (e.g. OMDB-GAP1_v1.1.tar.gz) from "
                "https://omdb.mathub.io/dataset and set datapath to this file"
            )

        super().__init__(
            dbpath=dbpath,
            subset=subset,
            load_only=load_only,
            collect_triples=collect_triples,
            available_properties=available_properties,
            units=units,
            environment_provider=environment_provider,
        )

        if download and not os.path.exists(dbpath):
            # Convert OMDB .tar.gz into a .db file
            self._convert()

    def _convert(self):
        """
        Converts .tar.gz to a .db file
        """
        print("Converting %s to a .db file.." % self.path)
        tar = tarfile.open(self.path, "r:gz")
        names = tar.getnames()
        tar.extractall()
        tar.close()

        structures = read("structures.xyz", index=":")
        Y = np.loadtxt("bandgaps.csv")
        [os.remove(name) for name in names]

        atoms_list = []
        property_list = []
        with connect(self.dbpath) as con:
            for i, at in enumerate(structures):
                atoms_list.append(at)
                property_list.append({OrganicMaterialsDatabase.BandGap: Y[i]})
        self.add_systems(atoms_list, property_list)



import os

from schnetpack.datasets import DownloadableAtomsData
from schnetpack.data.parsing import extxyz_to_db
from schnetpack.environment import SimpleEnvironmentProvider

__all__ = ["ExtXYZ"]


class ExtXYZ(DownloadableAtomsData):
    """
    Loader for MD data in extended XYZ format

    :param path: Path to database
    """

    E = "energy"
    F = "forces"

    def __init__(
        self,
        dbpath,
        xyzpath,
        subset=None,
        properties=None,
        environment_provider=SimpleEnvironmentProvider(),
        pair_provider=None,
        centering_function=True,
    ):
        available_properties = [ExtXYZ.E, ExtXYZ.F]
        units = [1.0, 1.0]

        if not os.path.exists(dbpath):
            os.makedirs(os.path.dirname(dbpath), exist_ok=True)
            extxyz_to_db(dbpath, xyzpath)

        super(ExtXYZ, self).__init__(
            dbpath=dbpath,
            subset=subset,
            load_only=properties,
            environment_provider=environment_provider,
            collect_triples=pair_provider,
            centering_function=centering_function,
            available_properties=available_properties,
            units=units,
        )



"""
Basic building blocks of SchNetPack models. Contains various basic and specialized network layers, layers for
cutoff functions, as well as several auxiliary layers and functions.
"""

from schnetpack.nn.acsf import *
from schnetpack.nn.activations import *
from schnetpack.nn.base import *
from schnetpack.nn.blocks import *
from schnetpack.nn.cfconv import *
from schnetpack.nn.cutoff import *
from schnetpack.nn.initializers import *
from schnetpack.nn.neighbors import *



import warnings

import numpy as np
import torch

from schnetpack import Properties


class Metric:
    r"""
    Base class for all metrics.

    Metrics measure the performance during the training and evaluation.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `MSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        self.target = target
        self.model_output = target if model_output is None else model_output
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name
        self.element_wise = element_wise

    def add_batch(self, batch, result):
        """Add a batch to calculate the metric on"""
        raise NotImplementedError

    def aggregate(self):
        """Aggregate metric over all previously added batches."""
        raise NotImplementedError

    def reset(self):
        """Reset metric attributes after aggregation to collect new batches."""
        pass


class ModelBias(Metric):
    r"""
    Calculates the bias of the model. For non-scalar quantities, the mean of
    all components is taken.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `MSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        name = "Bias_" + target if name is None else name
        super(ModelBias, self).__init__(
            target=target,
            model_output=model_output,
            name=name,
            element_wise=element_wise,
        )

        self.l2loss = 0.0
        self.n_entries = 0.0

    def reset(self):
        """Reset metric attributes after aggregation to collect new batches."""
        self.l2loss = 0.0
        self.n_entries = 0.0

    def _get_diff(self, y, yp):
        return y - yp

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        diff = self._get_diff(y, yp)
        self.l2loss += torch.sum(diff.view(-1)).detach().cpu().data.numpy()
        if self.element_wise:
            self.n_entries += (
                torch.sum(batch[Properties.atom_mask]).detach().cpu().data.numpy()
                * y.shape[-1]
            )
        else:
            self.n_entries += np.prod(y.shape)

    def aggregate(self):
        return self.l2loss / self.n_entries


class MeanSquaredError(Metric):
    r"""
    Metric for mean square error. For non-scalar quantities, the mean of all
    components is taken.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `MSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(
        self,
        target,
        model_output=None,
        bias_correction=None,
        name=None,
        element_wise=False,
    ):
        name = "MSE_" + target if name is None else name
        super(MeanSquaredError, self).__init__(
            target=target,
            model_output=model_output,
            name=name,
            element_wise=element_wise,
        )

        self.bias_correction = bias_correction

        self.l2loss = 0.0
        self.n_entries = 0.0

    def reset(self):
        """Reset metric attributes after aggregation to collect new batches."""
        self.l2loss = 0.0
        self.n_entries = 0.0

    def _get_diff(self, y, yp):
        diff = y - yp
        if self.bias_correction is not None:
            diff += self.bias_correction
        return diff

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        diff = self._get_diff(y, yp)
        self.l2loss += torch.sum(diff.view(-1) ** 2).detach().cpu().data.numpy()
        if self.element_wise:
            self.n_entries += (
                torch.sum(batch[Properties.atom_mask]).detach().cpu().data.numpy()
                * y.shape[-1]
            )
        else:
            self.n_entries += np.prod(y.shape)

    def aggregate(self):
        return self.l2loss / self.n_entries


class RootMeanSquaredError(MeanSquaredError):
    r"""
    Metric for root mean square error. For non-scalar quantities, the mean of
    all components is taken.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `RMSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(
        self,
        target,
        model_output=None,
        bias_correction=None,
        name=None,
        element_wise=False,
    ):
        name = "RMSE_" + target if name is None else name
        super(RootMeanSquaredError, self).__init__(
            target, model_output, bias_correction, name, element_wise=element_wise
        )

    def aggregate(self):
        """Aggregate metric over all previously added batches."""
        return np.sqrt(self.l2loss / self.n_entries)


class MeanAbsoluteError(Metric):
    r"""
    Metric for mean absolute error. For non-scalar quantities, the mean of all
    components is taken.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `MAE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(
        self,
        target,
        model_output=None,
        bias_correction=None,
        name=None,
        element_wise=False,
    ):
        name = "MAE_" + target if name is None else name
        super(MeanAbsoluteError, self).__init__(
            target=target,
            model_output=model_output,
            name=name,
            element_wise=element_wise,
        )

        self.bias_correction = bias_correction

        self.l1loss = 0.0
        self.n_entries = 0.0

    def reset(self):
        """Reset metric attributes after aggregation to collect new batches."""
        self.l1loss = 0.0
        self.n_entries = 0.0

    def _get_diff(self, y, yp):
        diff = y - yp
        if self.bias_correction is not None:
            diff += self.bias_correction
        return diff

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
                    # print(result.shape)
            else:
                result = result[self.model_output]
            yp = result

        # print(yp, yp.shape, y.shape)
        diff = self._get_diff(y, yp)
        # print(diff)
        # print()
        self.l1loss += (
            torch.sum(torch.abs(diff).view(-1), 0).detach().cpu().data.numpy()
        )
        if self.element_wise:
            self.n_entries += (
                torch.sum(batch[Properties.atom_mask]).detach().cpu().data.numpy()
                * y.shape[-1]
            )
        else:
            self.n_entries += np.prod(y.shape)

    def aggregate(self):
        """Aggregate metric over all previously added batches."""
        return self.l1loss / self.n_entries


class HeatmapMAE(MeanAbsoluteError):
    r"""
    Metric for heatmap of component-wise mean square error of non-scalar
    quantities.
    Do not use for datasets with different atomistic configurations!

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `HeatmapMAE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        name = "HeatmapMAE_" + target if name is None else name
        super(HeatmapMAE, self).__init__(
            target, model_output, name=name, element_wise=element_wise
        )

    def add_batch(self, batch, result):
        if self.element_wise and torch.sum(batch[Properties.atom_mask] == 0) != 0:
            warnings.warn(
                "MAEHeatmap should not be used for element-wise "
                + "properties with different sized molecules!"
            )

        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        diff = self._get_diff(y, yp)
        self.l1loss += torch.sum(torch.abs(diff), 0).detach().cpu().data.numpy()
        self.n_entries += y.size(0)

    def aggregate(self):
        """Aggregate metric over all previously added batches."""
        return self.l1loss / self.n_entries


class SumMAE(MeanAbsoluteError):
    r"""
    Metric for mean absolute error of length.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `LengthMAE_[target]` will be used (Default: None)
    """

    def __init__(
        self, target, model_output=None, axis=1, name=None, element_wise=False
    ):
        name = "SumMAE_" + target if name is None else name
        self.axis = axis
        super(SumMAE, self).__init__(
            target, model_output, name=name, element_wise=element_wise
        )

    def _get_diff(self, y, yp):
        ypl = torch.sum(yp, dim=self.axis)
        return torch.sum(torch.abs(y - ypl))


class LengthMSE(MeanSquaredError):
    r"""
    Metric for mean square error of length.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `LengthMSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        name = "LengthMSE_" + target if name is None else name
        super(LengthMSE, self).__init__(
            target, model_output, name=name, element_wise=element_wise
        )

    def _get_diff(self, y, yp):
        yl = torch.sqrt(torch.sum(y ** 2, dim=-1))
        ypl = torch.sqrt(torch.sum(yp ** 2, dim=-1))
        return torch.sum((yl - ypl) ** 2)


class LengthMAE(MeanAbsoluteError):
    r"""
    Metric for mean absolute error of length.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `LengthMAE_[target]` will be used (Default: None)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        name = "LengthMAE_" + target if name is None else name
        super(LengthMAE, self).__init__(
            target, model_output, name=name, element_wise=element_wise
        )

    def _get_diff(self, y, yp):
        yl = torch.sqrt(torch.sum(y ** 2, dim=-1))
        ypl = torch.sqrt(torch.sum(yp ** 2, dim=-1))
        return torch.sum(torch.abs(yl - ypl))


class LengthRMSE(RootMeanSquaredError):
    r"""
    Metric for root mean square error of length.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `LengthRMSE_[target]` will be used (Default: None)
        element_wise (bool): set to True if the model output is an element-wise
            property (forces, positions, ...)
    """

    def __init__(self, target, model_output=None, name=None, element_wise=False):
        name = "LengthRMSE_" + target if name is None else name
        super(LengthRMSE, self).__init__(
            target, model_output, name=name, element_wise=element_wise
        )

    def _get_diff(self, y, yp):
        yl = torch.sqrt(torch.sum(y ** 2, dim=-1))
        ypl = torch.sqrt(torch.sum(yp ** 2, dim=-1))
        return torch.sum((yl - ypl) ** 2)


class AngleMSE(MeanSquaredError):
    r"""
    Metric for mean square error of angles.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `AngleMSE_[target]` will be used (Default: None)
    """

    def __init__(self, target, model_output=None, name=None):
        name = "AngleMSE_" + target if name is None else name
        super(AngleMSE, self).__init__(target, model_output, name=name)

    def _get_diff(self, y, yp):
        y = y / torch.norm(y, dim=1, keepdim=True)
        yp = yp / torch.norm(yp, dim=1, keepdim=True)

        diff = torch.matmul(
            y.view(y.size(0), 1, y.size(1)), yp.view(y.size(0), y.size(1), 1)
        )[:, 0]
        diff = torch.clamp(diff, -1, 1)
        angle = torch.acos(diff)

        return angle

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        y = y.view(-1, y.size(-1))
        yp = yp.view(-1, yp.size(-1))

        diff = self._get_diff(y, yp)
        self.l2loss += torch.sum(diff ** 2).detach().cpu().data.numpy()
        self.n_entries += (
            torch.sum(torch.isnan(diff) == False).detach().cpu().data.numpy()
        )


class AngleMAE(MeanAbsoluteError):
    r"""
    Metric for mean absolute error of angles.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
                    `AngleMAE_[target]` will be used (Default: None)
    """

    def __init__(self, target, model_output=None, name=None):
        name = "AngleMAE_" + target if name is None else name
        super(AngleMAE, self).__init__(target, model_output, name=name)

    def _get_diff(self, y, yp):
        y = y / torch.norm(y, dim=1, keepdim=True)
        yp = yp / torch.norm(yp, dim=1, keepdim=True)

        diff = torch.matmul(
            y.view(y.size(0), 1, y.size(1)), yp.view(y.size(0), y.size(1), 1)
        )[:, 0]
        diff = torch.clamp(diff, -1, 1)
        angle = torch.acos(diff)

        return angle

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        y = y.view(-1, y.size(-1))
        yp = yp.view(-1, yp.size(-1))

        diff = self._get_diff(y, yp)
        self.l1loss += torch.sum(torch.abs(diff)).detach().cpu().data.numpy()
        self.n_entries += (
            torch.sum(torch.isnan(diff) == False).detach().cpu().data.numpy()
        )


class AngleRMSE(RootMeanSquaredError):
    r"""
    Metric for root mean square error of angles.

    Args:
        target (str): name of target property
        model_output (int, str): index or key, in case of multiple outputs
            (Default: None)
        name (str): name used in logging for this metric. If set to `None`,
            `AngleRMSE_[target]` will be used (Default: None)
    """

    def __init__(self, target, model_output=None, name=None):
        name = "AngleRMSE_" + target if name is None else name
        super(AngleRMSE, self).__init__(target, model_output, name=name)

    def _get_diff(self, y, yp):
        y = y / torch.norm(y, dim=1, keepdim=True)
        yp = yp / torch.norm(yp, dim=1, keepdim=True)

        diff = torch.matmul(
            y.view(y.size(0), 1, y.size(1)), yp.view(y.size(0), y.size(1), 1)
        )
        diff = torch.clamp(diff, -1, 1)
        angle = torch.acos(diff) / float(np.pi) * 180.0

        return angle

    def add_batch(self, batch, result):
        y = batch[self.target]
        if self.model_output is None:
            yp = result
        else:
            if type(self.model_output) is list:
                for idx in self.model_output:
                    result = result[idx]
            else:
                result = result[self.model_output]
            yp = result

        y = y.view(-1, y.size(-1))
        yp = yp.view(-1, yp.size(-1))

        diff = self._get_diff(y, yp)
        self.l2loss += torch.sum(diff ** 2).detach().cpu().data.numpy()
        self.n_entries += (
            torch.sum(torch.isnan(diff) == False).detach().cpu().data.numpy()
        )
